import os
from os.path import basename, realpath, relpath
from typing import Dict, Iterator, List, Optional, Set, Tuple

from elftools.common.exceptions import ELFError
from elftools.elf.elffile import ELFFile

from .lddtree import parse_ld_paths


def elf_read_dt_needed(fn: str) -> List[str]:
    needed = []
    with open(fn, "rb") as f:
        elf = ELFFile(f)
        section = elf.get_section_by_name(".dynamic")
        if section is None:
            raise ValueError("Could not find soname in %s" % fn)

        for t in section.iter_tags():
            if t.entry.d_tag == "DT_NEEDED":
                needed.append(t.needed)

    return needed


def elf_file_filter(paths: Iterator[str]) -> Iterator[Tuple[str, ELFFile]]:
    """Filter through an iterator of filenames and load up only ELF
    files
    """

    for path in paths:
        if path.endswith(".py"):
            continue
        else:
            try:
                with open(path, "rb") as f:
                    candidate = ELFFile(f)
                    yield path, candidate
            except ELFError:
                # not an elf file
                continue


def elf_find_versioned_symbols(elf: ELFFile) -> Iterator[Tuple[str, str]]:
    section = elf.get_section_by_name(".gnu.version_r")

    if section is not None:
        for verneed, verneed_iter in section.iter_versions():
            if verneed.name.startswith("ld-linux") or verneed.name in [
                "ld64.so.2",
                "ld64.so.1",
            ]:
                continue
            for vernaux in verneed_iter:
                yield (verneed.name, vernaux.name)


def elf_find_ucs2_symbols(elf: ELFFile) -> Iterator[str]:
    section = elf.get_section_by_name(".dynsym")
    if section is not None:
        # look for UCS2 symbols that are externally referenced
        for sym in section.iter_symbols():
            if (
                "PyUnicodeUCS2_" in sym.name
                and sym["st_shndx"] == "SHN_UNDEF"
                and sym["st_info"]["type"] == "STT_FUNC"
            ):

                yield sym.name


def elf_references_PyFPE_jbuf(elf: ELFFile) -> bool:
    offending_symbol_names = ("PyFPE_jbuf", "PyFPE_dummy", "PyFPE_counter")
    section = elf.get_section_by_name(".dynsym")
    if section is not None:
        # look for symbols that are externally referenced
        for sym in section.iter_symbols():
            if (
                sym.name in offending_symbol_names
                and sym["st_shndx"] == "SHN_UNDEF"
                and sym["st_info"]["type"] in ("STT_FUNC", "STT_NOTYPE")
            ):
                return True
    return False


def elf_is_python_extension(fn: str, elf: ELFFile) -> Tuple[bool, Optional[int]]:
    modname = basename(fn).split(".", 1)[0]
    module_init_f = {
        "init" + modname: 2,
        "PyInit_" + modname: 3,
        "_cffi_pypyinit_" + modname: 2,
    }

    sect = elf.get_section_by_name(".dynsym")
    if sect is None:
        return False, None

    for sym in sect.iter_symbols():
        if (
            sym.name in module_init_f
            and sym["st_shndx"] != "SHN_UNDEF"
            and sym["st_info"]["type"] == "STT_FUNC"
        ):

            return True, module_init_f[sym.name]

    return False, None


def elf_read_rpaths(fn: str) -> Dict[str, List[str]]:
    result = {"rpaths": [], "runpaths": []}  # type: Dict[str, List[str]]

    with open(fn, "rb") as f:
        elf = ELFFile(f)
        section = elf.get_section_by_name(".dynamic")
        if section is None:
            return result

        for t in section.iter_tags():
            if t.entry.d_tag == "DT_RPATH":
                result["rpaths"] = parse_ld_paths(t.rpath, root="/", path=fn)
            elif t.entry.d_tag == "DT_RUNPATH":
                result["runpaths"] = parse_ld_paths(t.runpath, root="/", path=fn)

    return result


def is_subdir(path: str, directory: str) -> bool:
    if path is None:
        return False

    path = realpath(path)
    directory = realpath(directory)

    relative = relpath(path, directory)
    if relative.startswith(os.pardir):
        return False
    return True


def get_undefined_symbols(path: str) -> Set[str]:
    undef_symbols = set()
    with open(path, "rb") as f:
        elf = ELFFile(f)
        section = elf.get_section_by_name(".dynsym")
        if section is not None:
            # look for all undef symbols
            for sym in section.iter_symbols():
                if sym["st_shndx"] == "SHN_UNDEF":
                    undef_symbols.add(sym.name)
    return undef_symbols


def filter_undefined_symbols(
    path: str, symbols: Dict[str, List[str]]
) -> Dict[str, List[str]]:
    if not symbols:
        return {}
    undef_symbols = set("*") | get_undefined_symbols(path)
    result = {}
    for lib, sym_list in symbols.items():
        intersection = set(sym_list) & undef_symbols
        if intersection:
            result[lib] = sorted(intersection)
    return result



import subprocess

from setuptools import setup

cmd = "gcc testpackage/testprogram.c -lgsl -lgslcblas -o testpackage/testprogram"
subprocess.check_call(cmd.split())

setup(
    name="testpackage",
    version="0.0.1",
    packages=["testpackage"],
    package_data={"testpackage": ["testprogram"]},
)



import itertools
import logging
import os
import platform
import re
import shutil
import stat
from collections import OrderedDict
from os.path import abspath, basename, dirname, exists, isabs
from os.path import join as pjoin
from subprocess import check_call
from typing import Dict, Iterable, List, Optional, Tuple

from auditwheel.patcher import ElfPatcher

from .elfutils import elf_read_dt_needed, elf_read_rpaths, is_subdir
from .hashfile import hashfile
from .policy import get_replace_platforms
from .wheel_abi import get_wheel_elfdata
from .wheeltools import InWheelCtx, add_platforms

logger = logging.getLogger(__name__)


# Copied from wheel 0.31.1
WHEEL_INFO_RE = re.compile(
    r"""^(?P<namever>(?P<name>.+?)-(?P<ver>\d.*?))(-(?P<build>\d.*?))?
     -(?P<pyver>[a-z].+?)-(?P<abi>.+?)-(?P<plat>.+?)(\.whl|\.dist-info)$""",
    re.VERBOSE,
).match


def repair_wheel(
    wheel_path: str,
    abis: List[str],
    lib_sdir: str,
    out_dir: str,
    update_tags: bool,
    patcher: ElfPatcher,
    exclude: List[str],
    strip: bool = False,
) -> Optional[str]:

    external_refs_by_fn = get_wheel_elfdata(wheel_path)[1]

    # Do not repair a pure wheel, i.e. has no external refs
    if not external_refs_by_fn:
        return None

    soname_map = {}  # type: Dict[str, Tuple[str, str]]
    if not isabs(out_dir):
        out_dir = abspath(out_dir)

    wheel_fname = basename(wheel_path)

    with InWheelCtx(wheel_path) as ctx:
        ctx.out_wheel = pjoin(out_dir, wheel_fname)

        match = WHEEL_INFO_RE(wheel_fname)
        if not match:
            raise ValueError("Failed to parse wheel file name: %s", wheel_fname)

        dest_dir = match.group("name") + lib_sdir

        if not exists(dest_dir):
            os.mkdir(dest_dir)

        # here, fn is a path to a python extension library in
        # the wheel, and v['libs'] contains its required libs
        for fn, v in external_refs_by_fn.items():
            ext_libs = v[abis[0]]["libs"]  # type: Dict[str, str]
            replacements = []  # type: List[Tuple[str, str]]
            for soname, src_path in ext_libs.items():

                if soname in exclude:
                    logger.info(f"Excluding {soname}")
                    continue

                if src_path is None:
                    raise ValueError(
                        (
                            "Cannot repair wheel, because required "
                            'library "%s" could not be located'
                        )
                        % soname
                    )

                new_soname, new_path = copylib(src_path, dest_dir, patcher)
                soname_map[soname] = (new_soname, new_path)
                replacements.append((soname, new_soname))
            if replacements:
                patcher.replace_needed(fn, *replacements)

            if len(ext_libs) > 0:
                new_rpath = os.path.relpath(dest_dir, os.path.dirname(fn))
                new_rpath = os.path.join("$ORIGIN", new_rpath)
                append_rpath_within_wheel(fn, new_rpath, ctx.name, patcher)

        # we grafted in a bunch of libraries and modified their sonames, but
        # they may have internal dependencies (DT_NEEDED) on one another, so
        # we need to update those records so each now knows about the new
        # name of the other.
        for old_soname, (new_soname, path) in soname_map.items():
            needed = elf_read_dt_needed(path)
            replacements = []
            for n in needed:
                if n in soname_map:
                    replacements.append((n, soname_map[n][0]))
            if replacements:
                patcher.replace_needed(path, *replacements)

        if update_tags:
            ctx.out_wheel = add_platforms(ctx, abis, get_replace_platforms(abis[0]))

        if strip:
            libs_to_strip = [path for (_, path) in soname_map.values()]
            extensions = external_refs_by_fn.keys()
            strip_symbols(itertools.chain(libs_to_strip, extensions))

    return ctx.out_wheel


def strip_symbols(libraries: Iterable[str]) -> None:
    for lib in libraries:
        logger.info("Stripping symbols from %s", lib)
        check_call(["strip", "-s", lib])


def copylib(src_path: str, dest_dir: str, patcher: ElfPatcher) -> Tuple[str, str]:
    """Graft a shared library from the system into the wheel and update the
    relevant links.

    1) Copy the file from src_path to dest_dir/
    2) Rename the shared object from soname to soname.<unique>
    3) If the library has a RUNPATH/RPATH, clear it and set RPATH to point to
    its new location.
    """
    # Copy the a shared library from the system (src_path) into the wheel
    # if the library has a RUNPATH/RPATH we clear it and set RPATH to point to
    # its new location.

    with open(src_path, "rb") as f:
        shorthash = hashfile(f)[:8]

    src_name = os.path.basename(src_path)
    base, ext = src_name.split(".", 1)
    if not base.endswith("-%s" % shorthash):
        new_soname = f"{base}-{shorthash}.{ext}"
    else:
        new_soname = src_name

    dest_path = os.path.join(dest_dir, new_soname)
    if os.path.exists(dest_path):
        return new_soname, dest_path

    logger.debug("Grafting: %s -> %s", src_path, dest_path)
    rpaths = elf_read_rpaths(src_path)
    shutil.copy2(src_path, dest_path)
    statinfo = os.stat(dest_path)
    if not statinfo.st_mode & stat.S_IWRITE:
        os.chmod(dest_path, statinfo.st_mode | stat.S_IWRITE)

    patcher.set_soname(dest_path, new_soname)

    if any(itertools.chain(rpaths["rpaths"], rpaths["runpaths"])):
        patcher.set_rpath(dest_path, dest_dir)

    return new_soname, dest_path


def append_rpath_within_wheel(
    lib_name: str, rpath: str, wheel_base_dir: str, patcher: ElfPatcher
) -> None:
    """Add a new rpath entry to a file while preserving as many existing
    rpath entries as possible.

    In order to preserve an rpath entry it must:

    1) Point to a location within wheel_base_dir.
    2) Not be a duplicate of an already-existing rpath entry.
    """
    if not isabs(lib_name):
        lib_name = abspath(lib_name)
    lib_dir = dirname(lib_name)
    if not isabs(wheel_base_dir):
        wheel_base_dir = abspath(wheel_base_dir)

    def is_valid_rpath(rpath: str) -> bool:
        return _is_valid_rpath(rpath, lib_dir, wheel_base_dir)

    old_rpaths = patcher.get_rpath(lib_name)
    rpaths = filter(is_valid_rpath, old_rpaths.split(":"))
    # Remove duplicates while preserving ordering
    # Fake an OrderedSet using OrderedDict
    rpath_set = OrderedDict([(old_rpath, "") for old_rpath in rpaths])
    rpath_set[rpath] = ""

    patcher.set_rpath(lib_name, ":".join(rpath_set))


def _is_valid_rpath(rpath: str, lib_dir: str, wheel_base_dir: str) -> bool:
    full_rpath_entry = _resolve_rpath_tokens(rpath, lib_dir)
    if not isabs(full_rpath_entry):
        logger.debug(
            f"rpath entry {rpath} could not be resolved to an "
            "absolute path -- discarding it."
        )
        return False
    elif not is_subdir(full_rpath_entry, wheel_base_dir):
        logger.debug(
            f"rpath entry {rpath} points outside the wheel -- " "discarding it."
        )
        return False
    else:
        logger.debug(f"Preserved rpath entry {rpath}")
        return True


def _resolve_rpath_tokens(rpath: str, lib_base_dir: str) -> str:
    # See https://www.man7.org/linux/man-pages/man8/ld.so.8.html#DESCRIPTION
    if platform.architecture()[0] == "64bit":
        system_lib_dir = "lib64"
    else:
        system_lib_dir = "lib"
    system_processor_type = platform.machine()
    token_replacements = {
        "ORIGIN": lib_base_dir,
        "LIB": system_lib_dir,
        "PLATFORM": system_processor_type,
    }
    for token, target in token_replacements.items():
        rpath = rpath.replace(f"${token}", target)  # $TOKEN
        rpath = rpath.replace(f"${{{token}}}", target)  # ${TOKEN}
    return rpath



import logging
import pathlib
import re
import subprocess
from typing import NamedTuple

from auditwheel.error import InvalidLibc

LOG = logging.getLogger(__name__)


class MuslVersion(NamedTuple):
    major: int
    minor: int
    patch: int


def find_musl_libc() -> pathlib.Path:
    try:
        (dl_path,) = list(pathlib.Path("/lib").glob("libc.musl-*.so.1"))
    except ValueError:
        LOG.debug("musl libc not detected")
        raise InvalidLibc

    return dl_path


def get_musl_version(ld_path: pathlib.Path) -> MuslVersion:
    try:
        ld = subprocess.run(
            [ld_path], check=False, errors="strict", stderr=subprocess.PIPE
        ).stderr
    except FileNotFoundError:
        LOG.error("Failed to determine musl version", exc_info=True)
        raise InvalidLibc

    match = re.search(
        r"Version " r"(?P<major>\d+)." r"(?P<minor>\d+)." r"(?P<patch>\d+)", ld
    )
    if not match:
        raise InvalidLibc

    return MuslVersion(
        int(match.group("major")), int(match.group("minor")), int(match.group("patch"))
    )



import logging
from os.path import abspath, basename, exists, join

logger = logging.getLogger(__name__)


def configure_parser(sub_parsers):
    help = "Add new platform ABI tags to a wheel"

    p = sub_parsers.add_parser("addtag", help=help, description=help)
    p.add_argument("WHEEL_FILE", help="Path to wheel file")
    p.add_argument(
        "-w",
        "--wheel-dir",
        dest="WHEEL_DIR",
        help=("Directory to store new wheel file (default:" ' "wheelhouse/")'),
        type=abspath,
        default="wheelhouse/",
    )
    p.set_defaults(func=execute)


def execute(args, p):
    import os

    from ._vendor.wheel.wheelfile import WHEEL_INFO_RE
    from .wheel_abi import NonPlatformWheel, analyze_wheel_abi
    from .wheeltools import InWheelCtx, WheelToolsError, add_platforms

    try:
        wheel_abi = analyze_wheel_abi(args.WHEEL_FILE)
    except NonPlatformWheel:
        logger.info(NonPlatformWheel.LOG_MESSAGE)
        return 1

    parsed_fname = WHEEL_INFO_RE.search(basename(args.WHEEL_FILE))
    in_fname_tags = parsed_fname.groupdict()["plat"].split(".")

    logger.info(
        '%s receives the following tag: "%s".',
        basename(args.WHEEL_FILE),
        wheel_abi.overall_tag,
    )
    logger.info("Use ``auditwheel show`` for more details")

    if wheel_abi.overall_tag in in_fname_tags:
        logger.info("No tags to be added. Exiting.")
        return 1

    # todo: move more of this logic to separate file
    if not exists(args.WHEEL_DIR):
        os.makedirs(args.WHEEL_DIR)

    with InWheelCtx(args.WHEEL_FILE) as ctx:
        try:
            out_wheel = add_platforms(ctx, [wheel_abi.overall_tag])
        except WheelToolsError as e:
            logger.exception("\n%s.", repr(e))
            return 1

        if out_wheel:
            # tell context manager to write wheel on exit with
            # the proper output directory
            ctx.out_wheel = join(args.WHEEL_DIR, basename(out_wheel))
            logger.info("\nUpdated wheel written to %s", out_wheel)
    return 0



import os

from auditwheel.tmpdirs import InGivenDirectory, InTemporaryDirectory


def test_intemporarydirectory():
    cwd = os.getcwd()
    with InTemporaryDirectory() as path:
        assert os.path.isdir(path)
        assert os.path.samefile(path, os.getcwd())
        assert not os.path.samefile(cwd, path)
    assert not os.path.exists(path)
    assert os.path.samefile(cwd, os.getcwd())


def test_intemporarydirectory_name():
    tmp_dir = InTemporaryDirectory()
    with tmp_dir as path:
        assert tmp_dir.name == path


def test_ingivendirectory(tmp_path):
    cwd = os.getcwd()
    expected_path = os.path.join(str(tmp_path), "foo")
    with InGivenDirectory(expected_path) as path:
        assert os.path.isdir(path)
        assert os.path.samefile(path, os.getcwd())
        assert os.path.samefile(path, expected_path)
    assert os.path.exists(path)
    assert os.path.samefile(cwd, os.getcwd())


def test_ingivendirectory_cwd(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    with InGivenDirectory() as path:
        assert os.path.isdir(path)
        assert os.path.samefile(path, os.getcwd())
        assert os.path.samefile(path, str(tmp_path))
    assert os.path.exists(path)


def test_ingivendirectory_name():
    given_dir = InGivenDirectory()
    with given_dir as path:
        assert given_dir.name == path



import pathlib
import subprocess

import pytest

HERE = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize("mode", ["repair", "addtag", "show"])
def test_non_platform_wheel_repair(mode):
    wheel = HERE / "plumbum-1.6.8-py2.py3-none-any.whl"
    proc = subprocess.run(
        ["auditwheel", mode, str(wheel)],
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    assert proc.returncode == 1
    assert "This does not look like a platform wheel" in proc.stderr
    assert "AttributeError" not in proc.stderr



from setuptools import Extension, setup

define_macros = [("_GNU_SOURCE", None)]
libraries = ["z", "c"]

setup(
    name="testzlib",
    version="0.0.1",
    ext_modules=[
        Extension(
            "testzlib",
            sources=["testzlib.c"],
            define_macros=define_macros,
            libraries=libraries,
        )
    ],
)



import functools
import itertools
import json
import logging
import os
from collections import defaultdict, namedtuple
from collections.abc import Mapping
from copy import deepcopy
from os.path import basename
from typing import Dict, Set

from .elfutils import (
    elf_file_filter,
    elf_find_ucs2_symbols,
    elf_find_versioned_symbols,
    elf_is_python_extension,
    elf_references_PyFPE_jbuf,
)
from .genericpkgctx import InGenericPkgCtx
from .lddtree import lddtree
from .policy import (
    POLICY_PRIORITY_HIGHEST,
    POLICY_PRIORITY_LOWEST,
    get_policy_name,
    lddtree_external_references,
    load_policies,
    versioned_symbols_policy,
)

log = logging.getLogger(__name__)
WheelAbIInfo = namedtuple(
    "WheelAbIInfo",
    [
        "overall_tag",
        "external_refs",
        "ref_tag",
        "versioned_symbols",
        "sym_tag",
        "ucs_tag",
        "pyfpe_tag",
        "blacklist_tag",
    ],
)


class WheelAbiError(Exception):
    """Root exception class"""


class NonPlatformWheel(WheelAbiError):
    """No ELF binaries in the wheel"""

    LOG_MESSAGE = (
        "This does not look like a platform wheel, no ELF executable "
        "or shared library file (including compiled Python C extension) "
        "found in the wheel archive"
    )


@functools.lru_cache()
def get_wheel_elfdata(wheel_fn: str):
    full_elftree = {}
    nonpy_elftree = {}
    full_external_refs = {}
    versioned_symbols = defaultdict(lambda: set())  # type: Dict[str, Set[str]]
    uses_ucs2_symbols = False
    uses_PyFPE_jbuf = False

    with InGenericPkgCtx(wheel_fn) as ctx:
        shared_libraries_in_purelib = []

        platform_wheel = False
        for fn, elf in elf_file_filter(ctx.iter_files()):
            platform_wheel = True

            # Check for invalid binary wheel format: no shared library should
            # be found in purelib
            so_path_split = fn.split(os.sep)

            # If this is in purelib, add it to the list of shared libraries in
            # purelib
            if "purelib" in so_path_split:
                shared_libraries_in_purelib.append(so_path_split[-1])

            # If at least one shared library exists in purelib, this is going
            # to fail and there's no need to do further checks
            if not shared_libraries_in_purelib:
                log.debug("processing: %s", fn)
                elftree = lddtree(fn)

                for key, value in elf_find_versioned_symbols(elf):
                    log.debug("key %s, value %s", key, value)
                    versioned_symbols[key].add(value)

                is_py_ext, py_ver = elf_is_python_extension(fn, elf)

                # If the ELF is a Python extention, we definitely need to
                # include its external dependencies.
                if is_py_ext:
                    full_elftree[fn] = elftree
                    uses_PyFPE_jbuf |= elf_references_PyFPE_jbuf(elf)
                    if py_ver == 2:
                        uses_ucs2_symbols |= any(
                            True for _ in elf_find_ucs2_symbols(elf)
                        )
                    full_external_refs[fn] = lddtree_external_references(
                        elftree, ctx.path
                    )
                else:
                    # If the ELF is not a Python extension, it might be
                    # included in the wheel already because auditwheel repair
                    # vendored it, so we will check whether we should include
                    # its internal references later.
                    nonpy_elftree[fn] = elftree

        if not platform_wheel:
            raise NonPlatformWheel

        # If at least one shared library exists in purelib, raise an error
        if shared_libraries_in_purelib:
            raise RuntimeError(
                (
                    "Invalid binary wheel, found the following shared "
                    "library/libraries in purelib folder:\n"
                    "\t%s\n"
                    "The wheel has to be platlib compliant in order to be "
                    "repaired by auditwheel."
                )
                % "\n\t".join(shared_libraries_in_purelib)
            )

        # Get a list of all external libraries needed by ELFs in the wheel.
        needed_libs = {
            lib
            for elf in itertools.chain(full_elftree.values(), nonpy_elftree.values())
            for lib in elf["needed"]
        }

        for fn in nonpy_elftree.keys():
            # If a non-pyextension ELF file is not needed by something else
            # inside the wheel, then it was not checked by the logic above and
            # we should walk its elftree.
            if basename(fn) not in needed_libs:
                full_elftree[fn] = nonpy_elftree[fn]

            # Even if a non-pyextension ELF file is not needed, we
            # should include it as an external reference, because
            # it might require additional external libraries.
            full_external_refs[fn] = lddtree_external_references(
                nonpy_elftree[fn], ctx.path
            )

    log.debug("full_elftree:\n%s", json.dumps(full_elftree, indent=4))
    log.debug(
        "full_external_refs (will be repaired):\n%s",
        json.dumps(full_external_refs, indent=4),
    )

    return (
        full_elftree,
        full_external_refs,
        versioned_symbols,
        uses_ucs2_symbols,
        uses_PyFPE_jbuf,
    )


def get_external_libs(external_refs) -> Dict[str, str]:
    """Get external library dependencies for all policies excluding the default
    linux policy
    :param external_refs: external references for all policies
    :return: {realpath: soname} e.g.
    {'/path/to/external_ref.so.1.2.3': 'external_ref.so.1'}
    """
    result: Dict[str, str] = {}
    for policy in external_refs.values():
        # linux tag (priority 0) has no white-list, do not analyze it
        if policy["priority"] == 0:
            continue
        # go through all libs, retrieving their soname and realpath
        for libname, realpath in policy["libs"].items():
            if realpath and realpath not in result.keys():
                result[realpath] = libname
    return result


def get_versioned_symbols(libs):
    """Get versioned symbols used in libraries
    :param libs: {realpath: soname} dict to search for versioned symbols e.g.
    {'/path/to/external_ref.so.1.2.3': 'external_ref.so.1'}
    :return: {soname: {depname: set([symbol_version])}} e.g.
    {'external_ref.so.1': {'libc.so.6', set(['GLIBC_2.5','GLIBC_2.12'])}}
    """
    result = {}
    for path, elf in elf_file_filter(libs.keys()):
        # {depname: set(symbol_version)}, e.g.
        # {'libc.so.6', set(['GLIBC_2.5','GLIBC_2.12'])}
        elf_versioned_symbols = defaultdict(lambda: set())
        for key, value in elf_find_versioned_symbols(elf):
            log.debug("path %s, key %s, value %s", path, key, value)
            elf_versioned_symbols[key].add(value)
        result[libs[path]] = elf_versioned_symbols
    return result


def get_symbol_policies(versioned_symbols, external_versioned_symbols, external_refs):
    """Get symbol policies
    Since white-list is different per policy, this function inspects
    versioned_symbol per policy when including external refs
    :param versioned_symbols: versioned symbols for the current wheel
    :param external_versioned_symbols: versioned symbols for external libs
    :param external_refs: external references for all policies
    :return: list of tuples of the form (policy_priority, versioned_symbols),
    e.g. [(100, {'libc.so.6', set(['GLIBC_2.5'])})]
    """
    result = []
    for policy in external_refs.values():
        # skip the linux policy
        if policy["priority"] == 0:
            continue
        policy_symbols = deepcopy(versioned_symbols)
        for soname in policy["libs"].keys():
            if soname not in external_versioned_symbols:
                continue
            ext_symbols = external_versioned_symbols[soname]
            for k in iter(ext_symbols):
                policy_symbols[k].update(ext_symbols[k])
        result.append((versioned_symbols_policy(policy_symbols), policy_symbols))
    return result


def analyze_wheel_abi(wheel_fn: str) -> WheelAbIInfo:
    external_refs = {
        p["name"]: {"libs": {}, "blacklist": {}, "priority": p["priority"]}
        for p in load_policies()
    }

    (
        elftree_by_fn,
        external_refs_by_fn,
        versioned_symbols,
        has_ucs2,
        uses_PyFPE_jbuf,
    ) = get_wheel_elfdata(wheel_fn)

    for fn in elftree_by_fn.keys():
        update(external_refs, external_refs_by_fn[fn])

    log.debug("external reference info")
    log.debug(json.dumps(external_refs, indent=4))

    external_libs = get_external_libs(external_refs)
    external_versioned_symbols = get_versioned_symbols(external_libs)
    symbol_policies = get_symbol_policies(
        versioned_symbols, external_versioned_symbols, external_refs
    )
    symbol_policy = versioned_symbols_policy(versioned_symbols)

    # let's keep the highest priority policy and
    # corresponding versioned_symbols
    symbol_policy, versioned_symbols = max(
        symbol_policies, key=lambda x: x[0], default=(symbol_policy, versioned_symbols)
    )

    ref_policy = max(
        (e["priority"] for e in external_refs.values() if len(e["libs"]) == 0),
        default=POLICY_PRIORITY_LOWEST,
    )

    blacklist_policy = max(
        (e["priority"] for e in external_refs.values() if len(e["blacklist"]) == 0),
        default=POLICY_PRIORITY_LOWEST,
    )

    if has_ucs2:
        ucs_policy = POLICY_PRIORITY_LOWEST
    else:
        ucs_policy = POLICY_PRIORITY_HIGHEST

    if uses_PyFPE_jbuf:
        pyfpe_policy = POLICY_PRIORITY_LOWEST
    else:
        pyfpe_policy = POLICY_PRIORITY_HIGHEST

    ref_tag = get_policy_name(ref_policy)
    sym_tag = get_policy_name(symbol_policy)
    ucs_tag = get_policy_name(ucs_policy)
    pyfpe_tag = get_policy_name(pyfpe_policy)
    blacklist_tag = get_policy_name(blacklist_policy)
    overall_tag = get_policy_name(
        min(symbol_policy, ref_policy, ucs_policy, pyfpe_policy, blacklist_policy)
    )

    return WheelAbIInfo(
        overall_tag,
        external_refs,
        ref_tag,
        versioned_symbols,
        sym_tag,
        ucs_tag,
        pyfpe_tag,
        blacklist_tag,
    )


def update(d, u):
    for k, v in u.items():
        if k == "blacklist":
            for lib, symbols in v.items():
                if lib not in d[k]:
                    d[k][lib] = list(symbols)
                else:
                    d[k][lib] = sorted(set(d[k][lib]) | set(symbols))
        elif isinstance(v, Mapping):
            r = update(d.get(k, {}), v)
            d[k] = r
        elif isinstance(v, (str, int, float, type(None))):
            d[k] = u[k]
        else:
            raise RuntimeError("!", d, k)
    return d



import glob
import io
import logging
import os
import os.path as op
import re
import shutil
import sys
import zipfile
from contextlib import contextmanager
from subprocess import CalledProcessError

import docker
import pytest
from elftools.elf.elffile import ELFFile

from auditwheel.policy import get_arch_name, get_priority_by_name

logger = logging.getLogger(__name__)


ENCODING = "utf-8"
PLATFORM = get_arch_name()
MANYLINUX1_IMAGE_ID = f"quay.io/pypa/manylinux1_{PLATFORM}:latest"
MANYLINUX2010_IMAGE_ID = f"quay.io/pypa/manylinux2010_{PLATFORM}:latest"
MANYLINUX2014_IMAGE_ID = f"quay.io/pypa/manylinux2014_{PLATFORM}:latest"
MANYLINUX_2_24_IMAGE_ID = f"quay.io/pypa/manylinux_2_24_{PLATFORM}:latest"
MANYLINUX_2_28_IMAGE_ID = f"quay.io/pypa/manylinux_2_28_{PLATFORM}:latest"
if PLATFORM in {"i686", "x86_64"}:
    MANYLINUX_IMAGES = {
        "manylinux_2_5": MANYLINUX1_IMAGE_ID,
        "manylinux_2_12": MANYLINUX2010_IMAGE_ID,
        "manylinux_2_17": MANYLINUX2014_IMAGE_ID,
        "manylinux_2_24": MANYLINUX_2_24_IMAGE_ID,
        "manylinux_2_28": MANYLINUX_2_28_IMAGE_ID,
    }
    POLICY_ALIASES = {
        "manylinux_2_5": ["manylinux1"],
        "manylinux_2_12": ["manylinux2010"],
        "manylinux_2_17": ["manylinux2014"],
    }
else:
    MANYLINUX_IMAGES = {
        "manylinux_2_17": MANYLINUX2014_IMAGE_ID,
        "manylinux_2_24": MANYLINUX_2_24_IMAGE_ID,
    }
    if PLATFORM in {"aarch64", "ppc64le"}:
        MANYLINUX_IMAGES["manylinux_2_28"] = MANYLINUX_2_28_IMAGE_ID
    POLICY_ALIASES = {
        "manylinux_2_17": ["manylinux2014"],
    }
DOCKER_CONTAINER_NAME = "auditwheel-test-anylinux"
PYTHON_MAJ_MIN = [str(i) for i in sys.version_info[:2]]
PYTHON_ABI_MAJ_MIN = "".join(PYTHON_MAJ_MIN)
PYTHON_ABI_FLAGS = "m" if sys.version_info.minor < 8 else ""
PYTHON_ABI = f"cp{PYTHON_ABI_MAJ_MIN}-cp{PYTHON_ABI_MAJ_MIN}{PYTHON_ABI_FLAGS}"
MANYLINUX_PYTHON_IMAGE_ID = f'python:{".".join(PYTHON_MAJ_MIN)}-slim'
MUSLLINUX_IMAGES = {
    "musllinux_1_1": f"quay.io/pypa/musllinux_1_1_{PLATFORM}:latest",
}
MUSLLINUX_PYTHON_IMAGE_ID = f'python:{".".join(PYTHON_MAJ_MIN)}-alpine'
DEVTOOLSET = {
    "manylinux_2_5": "devtoolset-2",
    "manylinux_2_12": "devtoolset-8",
    "manylinux_2_17": "devtoolset-10",
    "manylinux_2_24": "devtoolset-not-present",
    "manylinux_2_28": "gcc-toolset-11",
    "musllinux_1_1": "devtoolset-not-present",
}
PATH_DIRS = [
    f"/opt/python/{PYTHON_ABI}/bin",
    "/opt/rh/{devtoolset}/root/usr/bin",
    "/usr/local/sbin",
    "/usr/local/bin",
    "/usr/sbin",
    "/usr/bin",
    "/sbin",
    "/bin",
]
PATH = {k: ":".join(PATH_DIRS).format(devtoolset=v) for k, v in DEVTOOLSET.items()}
WHEEL_CACHE_FOLDER = op.expanduser("~/.cache/auditwheel_tests")
NUMPY_VERSION_MAP = {
    "36": "1.19.2",
    "37": "1.19.2",
    "38": "1.19.2",
    "39": "1.19.2",
    "310": "1.21.4",
}
NUMPY_VERSION = NUMPY_VERSION_MAP[PYTHON_ABI_MAJ_MIN]
ORIGINAL_NUMPY_WHEEL = f"numpy-{NUMPY_VERSION}-{PYTHON_ABI}-linux_{PLATFORM}.whl"
ORIGINAL_SIX_WHEEL = "six-1.11.0-py2.py3-none-any.whl"
SHOW_RE = re.compile(
    r'[\s](?P<wheel>\S+) is consistent with the following platform tag: "(?P<tag>\S+)"',
    flags=re.DOTALL,
)
TAG_RE = re.compile(r"^manylinux_(?P<major>[0-9]+)_(?P<minor>[0-9]+)_(?P<arch>\S+)$")


def find_src_folder():
    candidate = op.abspath(op.join(op.dirname(__file__), "../.."))
    contents = os.listdir(candidate)
    if "setup.py" in contents and "src" in contents:
        return candidate


def docker_start(image, volumes={}, env_variables={}):
    """Start a long waiting idle program in container

    Return the container object to be used for 'docker exec' commands.
    """
    # Make sure to use the latest public version of the docker image
    client = docker.from_env()

    dvolumes = {host: {"bind": ctr, "mode": "rw"} for (ctr, host) in volumes.items()}

    logger.info("Starting container with image %r", image)
    con = client.containers.run(
        image,
        ["sleep", "10000"],
        detach=True,
        volumes=dvolumes,
        environment=env_variables,
    )
    logger.info("Started container %s", con.id[:12])
    return con


@contextmanager
def docker_container_ctx(image, io_dir=None, env_variables={}):
    src_folder = find_src_folder()
    if src_folder is None:
        pytest.skip("Can only be run from the source folder")
    vols = {"/auditwheel_src": src_folder}
    if io_dir is not None:
        vols["/io"] = io_dir

    for key in env_variables:
        if key.startswith("COV_CORE_"):
            env_variables[key] = env_variables[key].replace(
                src_folder, "/auditwheel_src"
            )

    container = docker_start(image, vols, env_variables)
    try:
        yield container
    finally:
        container.remove(force=True)


def docker_exec(container, cmd, expected_retcode=0):
    logger.info("docker exec %s: %r", container.id[:12], cmd)
    ec, output = container.exec_run(cmd)
    output = output.decode(ENCODING)
    if ec != expected_retcode:
        print(output)
        raise CalledProcessError(ec, cmd, output=output)
    return output


@contextmanager
def tmp_docker_image(base, commands, setup_env={}):
    """Make a temporary docker image for tests

    Pulls the *base* image, runs *commands* inside it with *setup_env*, and
    commits the result as a new image. The image is removed on exiting the
    context.

    Making temporary images like this avoids each test having to re-run the
    same container setup steps.
    """
    with docker_container_ctx(base, env_variables=setup_env) as con:
        for cmd in commands:
            docker_exec(con, cmd)
        image = con.commit()

    logger.info("Made image %s based on %s", image.short_id, base)
    try:
        yield image.short_id
    finally:
        client = image.client
        client.images.remove(image.id)


def assert_show_output(manylinux_ctr, wheel, expected_tag, strict):
    output = docker_exec(manylinux_ctr, f"auditwheel show /io/{wheel}")
    output = output.replace("\n", " ")
    match = SHOW_RE.match(output)
    assert match
    assert match["wheel"] == wheel
    if strict or "musllinux" in expected_tag:
        assert match["tag"] == expected_tag
    else:
        expected_match = TAG_RE.match(expected_tag)
        assert expected_match, f"No match for tag {expected_tag}"
        expected_glibc = (int(expected_match["major"]), int(expected_match["minor"]))
        actual_match = TAG_RE.match(match["tag"])
        assert actual_match, f"No match for tag {match['tag']}"
        actual_glibc = (int(actual_match["major"]), int(actual_match["minor"]))
        assert expected_match["arch"] == actual_match["arch"]
        assert actual_glibc <= expected_glibc


def build_numpy(container, policy, output_dir):
    """Helper to build numpy from source using the specified container, into
    output_dir."""

    if policy.startswith("musllinux_"):
        docker_exec(container, "apk add openblas-dev")
    elif policy.startswith("manylinux_2_24_"):
        docker_exec(container, "apt-get update")
        docker_exec(
            container, "apt-get install -y --no-install-recommends libatlas-dev"
        )
    else:
        docker_exec(container, "yum install -y atlas atlas-devel")

    if op.exists(op.join(WHEEL_CACHE_FOLDER, policy, ORIGINAL_NUMPY_WHEEL)):
        # If numpy has already been built and put in cache, let's reuse this.
        shutil.copy2(
            op.join(WHEEL_CACHE_FOLDER, policy, ORIGINAL_NUMPY_WHEEL),
            op.join(output_dir, ORIGINAL_NUMPY_WHEEL),
        )
    else:
        # otherwise build the original linux_x86_64 numpy wheel from source
        # and put the result in the cache folder to speed-up future build.
        # This part of the build is independent of the auditwheel code-base
        # so it's safe to put it in cache.

        docker_exec(
            container,
            f"pip wheel -w /io --no-binary=:all: numpy=={NUMPY_VERSION}",
        )
        os.makedirs(op.join(WHEEL_CACHE_FOLDER, policy), exist_ok=True)
        shutil.copy2(
            op.join(output_dir, ORIGINAL_NUMPY_WHEEL),
            op.join(WHEEL_CACHE_FOLDER, policy, ORIGINAL_NUMPY_WHEEL),
        )
    orig_wheel, *_ = os.listdir(output_dir)
    return orig_wheel


class Anylinux:
    @pytest.fixture()
    def io_folder(self, tmp_path):
        d = tmp_path / "io"
        d.mkdir(exist_ok=True)
        return str(d)

    @pytest.fixture()
    def docker_python(self, docker_python_img, io_folder):
        with docker_container_ctx(docker_python_img, io_folder) as container:
            yield container

    @pytest.fixture()
    def any_manylinux_container(self, any_manylinux_img, io_folder):
        policy, manylinux_img = any_manylinux_img
        env = {"PATH": PATH[policy]}
        for key in os.environ:
            if key.startswith("COV_CORE_"):
                env[key] = os.environ[key]

        with docker_container_ctx(manylinux_img, io_folder, env) as container:
            platform_tag = ".".join(
                [f"{p}_{PLATFORM}" for p in [policy] + POLICY_ALIASES.get(policy, [])]
            )
            yield f"{policy}_{PLATFORM}", platform_tag, container

    def test_build_repair_numpy(
        self, any_manylinux_container, docker_python, io_folder
    ):
        # Integration test: repair numpy built from scratch
        policy, tag, manylinux_ctr = any_manylinux_container

        # First build numpy from source as a naive linux wheel that is tied
        # to system libraries (atlas, libgfortran...)
        orig_wheel = build_numpy(manylinux_ctr, policy, io_folder)
        assert orig_wheel == ORIGINAL_NUMPY_WHEEL
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the manylinux container
        repair_command = (
            f"auditwheel repair --plat {policy} --only-plat -w /io /io/{orig_wheel}"
        )
        docker_exec(manylinux_ctr, repair_command)
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"numpy-{NUMPY_VERSION}-{PYTHON_ABI}-{tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, policy, False)

        # Check that the repaired numpy wheel can be installed and executed
        # on a modern linux image.
        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        output = docker_exec(
            docker_python,
            "python /auditwheel_src/tests/integration/quick_check_numpy.py",
        )
        assert output.strip() == "ok"

        # Check that numpy f2py works with a more recent version of gfortran
        if policy.startswith("musllinux_"):
            docker_exec(docker_python, "apk add musl-dev gfortran")
        else:
            docker_exec(docker_python, "apt-get update -yqq")
            docker_exec(docker_python, "apt-get install -y gfortran")
        docker_exec(
            docker_python,
            "python -m numpy.f2py -c /auditwheel_src/tests/integration/foo.f90 -m foo",
        )

        # Check that the 2 fortran runtimes are well isolated and can be loaded
        # at once in the same Python program:
        docker_exec(docker_python, ["python", "-c", "'import numpy; import foo'"])

    @pytest.mark.skipif(
        PLATFORM != "x86_64", reason="Only needs checking on one platform"
    )
    def test_repair_exclude(self, any_manylinux_container, io_folder):
        """Test the --exclude argument to avoid grafting certain libraries."""

        policy, tag, manylinux_ctr = any_manylinux_container

        orig_wheel = build_numpy(manylinux_ctr, policy, io_folder)
        assert orig_wheel == ORIGINAL_NUMPY_WHEEL
        assert "manylinux" not in orig_wheel

        # Exclude libgfortran from grafting into the wheel
        excludes = {
            "manylinux_2_5_x86_64": ["libgfortran.so.1", "libgfortran.so.3"],
            "manylinux_2_12_x86_64": ["libgfortran.so.3", "libgfortran.so.5"],
            "manylinux_2_17_x86_64": ["libgfortran.so.3", "libgfortran.so.5"],
            "manylinux_2_24_x86_64": ["libgfortran.so.3"],
            "manylinux_2_28_x86_64": ["libgfortran.so.5"],
            "musllinux_1_1_x86_64": ["libgfortran.so.5"],
        }[policy]

        repair_command = [
            "auditwheel",
            "repair",
            "--plat",
            policy,
            "--only-plat",
            "-w",
            "/io",
        ]
        for exclude in excludes:
            repair_command.extend(["--exclude", exclude])
        repair_command.append(f"/io/{orig_wheel}")
        output = docker_exec(manylinux_ctr, repair_command)

        for exclude in excludes:
            assert f"Excluding {exclude}" in output
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"numpy-{NUMPY_VERSION}-{PYTHON_ABI}-{tag}.whl"
        assert repaired_wheel in filenames

        # Make sure we don't have libgfortran in the result
        contents = zipfile.ZipFile(os.path.join(io_folder, repaired_wheel)).namelist()
        assert not any(x for x in contents if "/libgfortran" in x)

    def test_build_wheel_with_binary_executable(
        self, any_manylinux_container, docker_python, io_folder
    ):
        # Test building a wheel that contains a binary executable (e.g., a program)

        policy, tag, manylinux_ctr = any_manylinux_container
        if policy.startswith("musllinux_"):
            docker_exec(manylinux_ctr, "apk add gsl-dev")
        elif policy.startswith("manylinux_2_24_"):
            docker_exec(manylinux_ctr, "apt-get update")
            docker_exec(
                manylinux_ctr, "apt-get install -y --no-install-recommends libgsl-dev"
            )
        else:
            docker_exec(manylinux_ctr, "yum install -y gsl-devel")

        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/testpackage && "
                "if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi && "
                "python -m pip wheel --no-deps -w /io .",
            ],
        )

        filenames = os.listdir(io_folder)
        assert filenames == ["testpackage-0.0.1-py3-none-any.whl"]
        orig_wheel = filenames[0]
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the appropriate manylinux container
        repair_command = (
            f"auditwheel repair --plat {policy} --only-plat -w /io /io/{orig_wheel}"
        )
        docker_exec(manylinux_ctr, repair_command)
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"testpackage-0.0.1-py3-none-{tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, policy, False)

        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        output = docker_exec(
            docker_python,
            ["python", "-c", "from testpackage import runit; print(runit(1.5))"],
        )
        assert output.strip() == "2.25"

    def test_build_repair_pure_wheel(self, any_manylinux_container, io_folder):
        policy, tag, manylinux_ctr = any_manylinux_container

        if op.exists(op.join(WHEEL_CACHE_FOLDER, policy, ORIGINAL_SIX_WHEEL)):
            # If six has already been built and put in cache, let's reuse this.
            shutil.copy2(
                op.join(WHEEL_CACHE_FOLDER, policy, ORIGINAL_SIX_WHEEL),
                op.join(io_folder, ORIGINAL_SIX_WHEEL),
            )
            logger.info(f"Copied six wheel from {WHEEL_CACHE_FOLDER} to {io_folder}")
        else:
            docker_exec(manylinux_ctr, "pip wheel -w /io --no-binary=:all: six==1.11.0")
            os.makedirs(op.join(WHEEL_CACHE_FOLDER, policy), exist_ok=True)
            shutil.copy2(
                op.join(io_folder, ORIGINAL_SIX_WHEEL),
                op.join(WHEEL_CACHE_FOLDER, policy, ORIGINAL_SIX_WHEEL),
            )

        filenames = os.listdir(io_folder)
        assert filenames == [ORIGINAL_SIX_WHEEL]
        orig_wheel = filenames[0]
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the manylinux container
        repair_command = f"auditwheel repair --plat {policy} -w /io /io/{orig_wheel}"
        output = docker_exec(manylinux_ctr, repair_command, expected_retcode=1)
        assert "This does not look like a platform wheel" in output

        output = docker_exec(
            manylinux_ctr, f"auditwheel show /io/{orig_wheel}", expected_retcode=1
        )
        assert "This does not look like a platform wheel" in output

    @pytest.mark.parametrize("dtag", ["rpath", "runpath"])
    def test_build_wheel_depending_on_library_with_rpath(
        self, any_manylinux_container, docker_python, io_folder, dtag
    ):
        # Test building a wheel that contains an extension depending on a library
        # with RPATH or RUNPATH set.
        # Following checks are performed:
        # - check if RUNPATH is replaced by RPATH
        # - check if RPATH location is correct, i.e. it is inside .libs directory
        #   where all gathered libraries are put

        policy, tag, manylinux_ctr = any_manylinux_container

        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                (
                    "cd /auditwheel_src/tests/integration/testrpath &&"
                    "if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi && "
                    f"DTAG={dtag} python setup.py bdist_wheel -d /io"
                ),
            ],
        )
        with open(
            op.join(op.dirname(__file__), "testrpath", "a", "liba.so"), "rb"
        ) as f:
            elf = ELFFile(f)
            dynamic = elf.get_section_by_name(".dynamic")
            tags = {t.entry.d_tag for t in dynamic.iter_tags()}
            assert f"DT_{dtag.upper()}" in tags
        filenames = os.listdir(io_folder)
        assert filenames == [f"testrpath-0.0.1-{PYTHON_ABI}-linux_{PLATFORM}.whl"]
        orig_wheel = filenames[0]
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the appropriate manylinux container
        repair_command = (
            f"auditwheel repair --plat {policy} --only-plat -w /io /io/{orig_wheel}"
        )
        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "LD_LIBRARY_PATH="
                "/auditwheel_src/tests/integration/testrpath/a:$LD_LIBRARY_PATH "
                + repair_command,
            ],
        )
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"testrpath-0.0.1-{PYTHON_ABI}-{tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, policy, False)

        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        output = docker_exec(
            docker_python,
            [
                "python",
                "-c",
                "from testrpath import testrpath; print(testrpath.func())",
            ],
        )
        assert output.strip() == "11"
        with zipfile.ZipFile(os.path.join(io_folder, repaired_wheel)) as w:
            for name in w.namelist():
                if "testrpath/.libs/lib" in name:
                    with w.open(name) as f:
                        elf = ELFFile(io.BytesIO(f.read()))
                        dynamic = elf.get_section_by_name(".dynamic")
                        assert (
                            len(
                                [
                                    t
                                    for t in dynamic.iter_tags()
                                    if t.entry.d_tag == "DT_RUNPATH"
                                ]
                            )
                            == 0
                        )
                        if ".libs/liba" in name:
                            rpath_tags = [
                                t
                                for t in dynamic.iter_tags()
                                if t.entry.d_tag == "DT_RPATH"
                            ]
                            assert len(rpath_tags) == 1
                            assert rpath_tags[0].rpath == "$ORIGIN/."

    def test_build_repair_multiple_top_level_modules_wheel(
        self, any_manylinux_container, docker_python, io_folder
    ):

        policy, tag, manylinux_ctr = any_manylinux_container

        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/multiple_top_level && "
                "if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi && "
                "make clean all && "
                "pip wheel . -w /io",
            ],
        )

        filenames = os.listdir(io_folder)
        assert filenames == [
            f"multiple_top_level-1.0-{PYTHON_ABI}-linux_{PLATFORM}.whl"
        ]
        orig_wheel = filenames[0]
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the appropriate manylinux container
        repair_command = (
            f"auditwheel repair --plat {policy} --only-plat -w /io /io/{orig_wheel}"
        )
        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                (
                    "LD_LIBRARY_PATH="
                    "/auditwheel_src/tests/integration/multiple_top_level/lib-src/a:"
                    "/auditwheel_src/tests/integration/multiple_top_level/lib-src/b:"
                    "$LD_LIBRARY_PATH "
                )
                + repair_command,
            ],
        )
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"multiple_top_level-1.0-{PYTHON_ABI}-{tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, policy, False)

        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        for mod, func, expected in [
            ("example_a", "example_a", "11"),
            ("example_b", "example_b", "110"),
        ]:
            output = docker_exec(
                docker_python,
                [
                    "python",
                    "-c",
                    f"from {mod} import {func}; print({func}())",
                ],
            ).strip()
            assert output.strip() == expected
        with zipfile.ZipFile(os.path.join(io_folder, repaired_wheel)) as w:
            for lib_name in ["liba", "libb"]:
                assert any(
                    re.match(rf"multiple_top_level.libs/{lib_name}.*\.so", name)
                    for name in w.namelist()
                )

    def test_build_repair_wheel_with_internal_rpath(
        self, any_manylinux_container, docker_python, io_folder
    ):
        policy, tag, manylinux_ctr = any_manylinux_container

        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/internal_rpath && "
                "if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi && "
                "make clean all && "
                "mv lib-src/a/liba.so internal_rpath && "
                "pip wheel . -w /io",
            ],
        )

        filenames = os.listdir(io_folder)
        assert filenames == [f"internal_rpath-1.0-{PYTHON_ABI}-linux_{PLATFORM}.whl"]
        orig_wheel = filenames[0]
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the appropriate manylinux container
        repair_command = (
            f"auditwheel repair --plat {policy} --only-plat -w /io /io/{orig_wheel}"
        )
        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                (
                    "LD_LIBRARY_PATH="
                    "/auditwheel_src/tests/integration/internal_rpath/lib-src/b:"
                    "$LD_LIBRARY_PATH "
                )
                + repair_command,
            ],
        )
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"internal_rpath-1.0-{PYTHON_ABI}-{tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, policy, False)

        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        for mod, func, expected in [
            ("example_a", "example_a", "11"),
            ("example_b", "example_b", "10"),
        ]:
            output = docker_exec(
                docker_python,
                [
                    "python",
                    "-c",
                    f"from internal_rpath.{mod} import {func}; print({func}())",
                ],
            ).strip()
            assert output.strip() == expected
        with zipfile.ZipFile(os.path.join(io_folder, repaired_wheel)) as w:
            for lib_name in ["libb"]:
                assert any(
                    re.match(rf"internal_rpath.libs/{lib_name}.*\.so", name)
                    for name in w.namelist()
                )

    def test_strip_wheel(self, any_manylinux_container, docker_python, io_folder):
        policy, tag, manylinux_ctr = any_manylinux_container
        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/sample_extension && "
                "if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi && "
                "python -m pip wheel --no-deps -w /io .",
            ],
        )

        orig_wheel, *_ = os.listdir(io_folder)
        assert orig_wheel.startswith("sample_extension-0.1.0")

        # Repair the wheel using the appropriate manylinux container
        repair_command = (
            f"auditwheel repair --plat {policy} --strip -w /io /io/{orig_wheel}"
        )
        docker_exec(manylinux_ctr, repair_command)

        repaired_wheel, *_ = glob.glob(f"{io_folder}/*{policy}*.whl")
        repaired_wheel = os.path.basename(repaired_wheel)

        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        output = docker_exec(
            docker_python,
            [
                "python",
                "-c",
                "from sample_extension import test_func; print(test_func(1))",
            ],
        )
        assert output.strip() == "2"

    def test_nonpy_rpath(self, any_manylinux_container, docker_python, io_folder):
        # Tests https://github.com/pypa/auditwheel/issues/136
        policy, tag, manylinux_ctr = any_manylinux_container
        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/nonpy_rpath && "
                "if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi && "
                "python -m pip wheel --no-deps -w /io .",
            ],
        )

        orig_wheel, *_ = os.listdir(io_folder)
        assert orig_wheel.startswith("nonpy_rpath-0.1.0")
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the appropriate manylinux container
        repair_command = (
            f"auditwheel repair --plat {policy} --only-plat -w /io /io/{orig_wheel}"
        )
        docker_exec(manylinux_ctr, ["bash", "-c", repair_command])
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"nonpy_rpath-0.1.0-{PYTHON_ABI}-{tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, policy, False)

        # Test the resulting wheel outside the manylinux container
        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        docker_exec(
            docker_python,
            [
                "python",
                "-c",
                "import nonpy_rpath\n"
                "assert nonpy_rpath.crypt_something().startswith('*')",
            ],
        )

    def test_glibcxx_3_4_25(self, any_manylinux_container, docker_python, io_folder):
        policy, tag, manylinux_ctr = any_manylinux_container
        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/test_glibcxx_3_4_25 && "
                "if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi && "
                "python -m pip wheel --no-deps -w /io .",
            ],
        )

        orig_wheel, *_ = os.listdir(io_folder)
        assert orig_wheel.startswith("testentropy-0.0.1")

        # Repair the wheel using the appropriate manylinux container
        repair_command = f"auditwheel repair --plat {policy} -w /io /io/{orig_wheel}"
        if policy.startswith("manylinux_2_28_"):
            with pytest.raises(CalledProcessError):
                docker_exec(manylinux_ctr, repair_command)
            # TODO if a "permissive" mode is implemented, add the relevant flag to the
            # repair_command here and drop the return statement below
            return

        docker_exec(manylinux_ctr, repair_command)

        repaired_wheel, *_ = glob.glob(f"{io_folder}/*{policy}*.whl")
        repaired_wheel = os.path.basename(repaired_wheel)

        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        docker_exec(
            docker_python,
            [
                "python",
                "-c",
                "from testentropy import run; exit(run())",
            ],
        )


class TestManylinux(Anylinux):
    @pytest.fixture(scope="session")
    def docker_python_img(self):
        """The glibc Python base image with up-to-date pip"""
        with tmp_docker_image(
            MANYLINUX_PYTHON_IMAGE_ID, ["pip install -U pip"]
        ) as img_id:
            yield img_id

    @pytest.fixture(scope="session", params=MANYLINUX_IMAGES.keys())
    def any_manylinux_img(self, request):
        """Each manylinux image, with auditwheel installed.

        Plus up-to-date pip, setuptools and pytest-cov
        """
        policy = request.param
        if policy == "manylinux_2_5" and PYTHON_ABI_MAJ_MIN in {"310"}:
            pytest.skip(f"manylinux_2_5 images do not support cp{PYTHON_ABI_MAJ_MIN}")
        base = MANYLINUX_IMAGES[policy]
        env = {"PATH": PATH[policy]}
        with tmp_docker_image(
            base,
            [
                'git config --global --add safe.directory "/auditwheel_src"',
                "pip install -U pip setuptools pytest-cov",
                "pip install -U -e /auditwheel_src",
            ],
            env,
        ) as img_id:
            yield policy, img_id

    @pytest.mark.parametrize("with_dependency", ["0", "1"])
    def test_build_wheel_with_image_dependencies(
        self, with_dependency, any_manylinux_container, docker_python, io_folder
    ):
        # try to repair the wheel targeting different policies
        #
        # with_dependency == 0
        #   The python module has no dependencies that should be grafted-in and
        #   uses versioned symbols not available on policies pre-dating the policy
        #   matching the image being tested.
        # with_dependency == 1
        #   The python module itself does not use versioned symbols but has a
        #   dependency that should be grafted-in that uses versioned symbols not
        #   available on policies pre-dating the policy matching the image being
        #   tested.

        policy, tag, manylinux_ctr = any_manylinux_container

        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/testdependencies && "
                f"WITH_DEPENDENCY={with_dependency} python setup.py -v build_ext -f "
                "bdist_wheel -d /io",
            ],
        )

        filenames = os.listdir(io_folder)
        orig_wheel = filenames[0]
        assert "manylinux" not in orig_wheel

        repair_command = (
            "LD_LIBRARY_PATH="
            "/auditwheel_src/tests/integration/testdependencies:$LD_LIBRARY_PATH "
            "auditwheel -v repair --plat {policy} -w /io /io/{orig_wheel}"
        )

        policy_priority = get_priority_by_name(policy)
        older_policies = [
            f"{p}_{PLATFORM}"
            for p in MANYLINUX_IMAGES.keys()
            if policy_priority < get_priority_by_name(f"{p}_{PLATFORM}")
        ]
        for target_policy in older_policies:
            # we shall fail to repair the wheel when targeting an older policy than
            # the one matching the image
            with pytest.raises(CalledProcessError):
                docker_exec(
                    manylinux_ctr,
                    [
                        "bash",
                        "-c",
                        repair_command.format(
                            policy=target_policy, orig_wheel=orig_wheel
                        ),
                    ],
                )

        # check all works properly when targeting the policy matching the image
        docker_exec(
            manylinux_ctr,
            ["bash", "-c", repair_command.format(policy=policy, orig_wheel=orig_wheel)],
        )
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        repaired_wheel = f"testdependencies-0.0.1-{PYTHON_ABI}-{tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, policy, True)

        # check the original wheel with a dependency was not compliant
        # and check the one without a dependency was already compliant
        if with_dependency == "1":
            expected = f"linux_{PLATFORM}"
        else:
            expected = policy
        assert_show_output(manylinux_ctr, orig_wheel, expected, True)

        docker_exec(docker_python, "pip install /io/" + repaired_wheel)
        docker_exec(
            docker_python,
            [
                "python",
                "-c",
                "from sys import exit; from testdependencies import run; exit(run())",
            ],
        )

    @pytest.mark.parametrize(
        "target_policy",
        [f"{p}_{PLATFORM}" for p in list(MANYLINUX_IMAGES.keys())]
        + [f"{p}_{PLATFORM}" for aliases in POLICY_ALIASES.values() for p in aliases],
    )
    @pytest.mark.parametrize("only_plat", [True, False])
    def test_build_wheel_compat(
        self,
        target_policy,
        only_plat,
        any_manylinux_container,
        docker_python,
        io_folder,
    ):
        # test building wheels with compatibility with older spec
        # check aliases for older spec

        policy, tag, manylinux_ctr = any_manylinux_container

        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/testsimple "
                "&& python -m pip wheel --no-deps -w /io .",
            ],
        )

        filenames = os.listdir(io_folder)
        orig_wheel = filenames[0]
        assert "manylinux" not in orig_wheel

        if PLATFORM in {"x86_64", "i686"}:
            expect = f"manylinux_2_5_{PLATFORM}"
            expect_tag = f"manylinux_2_5_{PLATFORM}.manylinux1_{PLATFORM}"
        else:
            expect = f"manylinux_2_17_{PLATFORM}"
            expect_tag = f"manylinux_2_17_{PLATFORM}.manylinux2014_{PLATFORM}"

        target_tag = target_policy
        for pep600_policy, aliases in POLICY_ALIASES.items():
            policy_ = f"{pep600_policy}_{PLATFORM}"
            aliases_ = [f"{p}_{PLATFORM}" for p in aliases]
            if target_policy == policy_ or target_policy in aliases_:
                target_tag = f'{policy_}.{".".join(aliases_)}'

        only_plat_arg = "--only-plat" if only_plat else ""
        # we shall ba able to repair the wheel for all targets
        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                f"auditwheel -v repair --plat {target_policy} {only_plat_arg} -w /io"
                f" /io/{orig_wheel}",
            ],
        )
        filenames = os.listdir(io_folder)
        assert len(filenames) == 2
        if only_plat or target_tag == expect_tag:
            repaired_tag = target_tag
        else:
            repaired_tag = f"{expect_tag}.{target_tag}"
        repaired_wheel = f"testsimple-0.0.1-{PYTHON_ABI}-{repaired_tag}.whl"
        assert repaired_wheel in filenames
        assert_show_output(manylinux_ctr, repaired_wheel, expect, True)

        docker_exec(docker_python, f"pip install /io/{repaired_wheel}")
        docker_exec(
            docker_python,
            [
                "python",
                "-c",
                "from sys import exit; from testsimple import run; exit(run())",
            ],
        )

    def test_zlib_blacklist(self, any_manylinux_container, docker_python, io_folder):
        policy, tag, manylinux_ctr = any_manylinux_container
        if policy.startswith("manylinux_2_17_"):
            pytest.skip(f"{policy} image has no blacklist symbols in libz.so.1")

        docker_exec(
            manylinux_ctr,
            [
                "bash",
                "-c",
                "cd /auditwheel_src/tests/integration/testzlib "
                "&& python -m pip wheel --no-deps -w /io .",
            ],
        )

        orig_wheel, *_ = os.listdir(io_folder)
        assert orig_wheel.startswith("testzlib-0.0.1")

        # Repair the wheel using the appropriate manylinux container
        repair_command = f"auditwheel repair --plat {policy} -w /io /io/{orig_wheel}"
        with pytest.raises(CalledProcessError):
            docker_exec(manylinux_ctr, repair_command)

        # Check auditwheel show warns about the black listed symbols
        output = docker_exec(manylinux_ctr, f"auditwheel -v show /io/{orig_wheel}")
        assert "black-listed symbol dependencies" in output.replace("\n", " ")


class TestMusllinux(Anylinux):
    @pytest.fixture(scope="session")
    def docker_python_img(self):
        """The alpine Python base image with up-to-date pip"""
        with tmp_docker_image(
            MUSLLINUX_PYTHON_IMAGE_ID, ["pip install -U pip"]
        ) as img_id:
            yield img_id

    @pytest.fixture(scope="session", params=MUSLLINUX_IMAGES.keys())
    def any_manylinux_img(self, request):
        """Each musllinux image, with auditwheel installed.

        Plus up-to-date pip, setuptools and pytest-cov
        """
        policy = request.param
        base = MUSLLINUX_IMAGES[policy]
        env = {"PATH": PATH[policy]}
        with tmp_docker_image(
            base,
            [
                'git config --global --add safe.directory "/auditwheel_src"',
                "pip install -U pip setuptools pytest-cov",
                "pip install -U -e /auditwheel_src",
            ],
            env,
        ) as img_id:
            yield policy, img_id



import logging
from typing import Dict, List, Set

from . import load_policies

log = logging.getLogger(__name__)


def versioned_symbols_policy(versioned_symbols: Dict[str, Set[str]]) -> int:
    def policy_is_satisfied(
        policy_name: str, policy_sym_vers: Dict[str, Set[str]]
    ) -> bool:
        policy_satisfied = True
        for name in set(required_vers) & set(policy_sym_vers):
            if not required_vers[name].issubset(policy_sym_vers[name]):
                for symbol in required_vers[name] - policy_sym_vers[name]:
                    log.debug(
                        "Package requires %s, incompatible with "
                        "policy %s which requires %s",
                        symbol,
                        policy_name,
                        policy_sym_vers[name],
                    )
                policy_satisfied = False
        return policy_satisfied

    required_vers = {}  # type: Dict[str, Set[str]]
    for symbols in versioned_symbols.values():
        for symbol in symbols:
            sym_name, _, _ = symbol.partition("_")
            required_vers.setdefault(sym_name, set()).add(symbol)
    matching_policies = []  # type: List[int]
    for p in load_policies():
        policy_sym_vers = {
            sym_name: {sym_name + "_" + version for version in versions}
            for sym_name, versions in p["symbol_versions"].items()
        }
        if policy_is_satisfied(p["name"], policy_sym_vers):
            matching_policies.append(p["priority"])

    if len(matching_policies) == 0:
        # the base policy (generic linux) should always match
        raise RuntimeError("Internal error")

    return max(matching_policies)



import subprocess
from unittest.mock import patch

import pytest

from auditwheel.error import InvalidLibc
from auditwheel.musllinux import find_musl_libc, get_musl_version


@patch("auditwheel.musllinux.pathlib.Path")
def test_find_musllinux_not_found(path_mock):
    path_mock.return_value.glob.return_value = []
    with pytest.raises(InvalidLibc):
        find_musl_libc()


@patch("auditwheel.musllinux.pathlib.Path")
def test_find_musllinux_found(path_mock):
    path_mock.return_value.glob.return_value = ["/lib/ld-musl-x86_64.so.1"]
    musl = find_musl_libc()
    assert musl


def test_get_musl_version_invalid_path():
    with pytest.raises(InvalidLibc):
        get_musl_version("/tmp/no/executable/here")


@patch("auditwheel.musllinux.subprocess.run")
def test_get_musl_version_invalid_version(run_mock):
    run_mock.return_value = subprocess.CompletedProcess([], 1, None, "Version 1.1")
    with pytest.raises(InvalidLibc):
        get_musl_version("anything")


@patch("auditwheel.musllinux.subprocess.run")
def test_get_musl_version_valid_version(run_mock):
    run_mock.return_value = subprocess.CompletedProcess([], 1, None, "Version 5.6.7")
    version = get_musl_version("anything")
    assert version.major == 5
    assert version.minor == 6
    assert version.patch == 7



from setuptools import Extension, find_packages, setup

package_name = "internal_rpath"
setup(
    name=package_name,
    version="1.0",
    description="Auditwheel multiple top-level extensions example",
    package_data={package_name: ["liba.so"]},
    packages=find_packages(),
    ext_modules=[
        Extension(
            f"{package_name}.example_a",
            ["src/example_a.pyx"],
            include_dirs=["lib-src/a"],
            library_dirs=[package_name],
            libraries=["a"],
            extra_link_args=["-Wl,-rpath,$ORIGIN"],
        ),
        Extension(
            f"{package_name}.example_b",
            ["src/example_b.pyx"],
            include_dirs=["lib-src/b"],
            library_dirs=["lib-src/b"],
            libraries=["b"],
        ),
    ],
)



import json
import logging
import platform as _platform_module
import sys
from collections import defaultdict
from os.path import abspath, dirname, join
from pathlib import Path
from typing import Dict, List, Optional, Set

from ..libc import Libc, get_libc
from ..musllinux import find_musl_libc, get_musl_version

_HERE = Path(__file__).parent

logger = logging.getLogger(__name__)

# https://docs.python.org/3/library/platform.html#platform.architecture
bits = 8 * (8 if sys.maxsize > 2**32 else 4)


def get_arch_name() -> str:
    machine = _platform_module.machine()
    if machine not in {"x86_64", "i686"}:
        return machine
    else:
        return {64: "x86_64", 32: "i686"}[bits]


_ARCH_NAME = get_arch_name()
_LIBC = get_libc()


def _validate_pep600_compliance(policies) -> None:
    symbol_versions: Dict[str, Dict[str, Set[str]]] = {}
    lib_whitelist: Set[str] = set()
    for policy in sorted(policies, key=lambda x: x["priority"], reverse=True):
        if policy["name"] == "linux":
            continue
        if not lib_whitelist.issubset(set(policy["lib_whitelist"])):
            diff = lib_whitelist - set(policy["lib_whitelist"])
            raise ValueError(
                'Invalid "policy.json" file. Missing whitelist libraries in '
                f'"{policy["name"]}" compared to previous policies: {diff}'
            )
        lib_whitelist.update(policy["lib_whitelist"])
        for arch in policy["symbol_versions"].keys():
            symbol_versions_arch = symbol_versions.get(arch, defaultdict(set))
            for prefix in policy["symbol_versions"][arch].keys():
                policy_symbol_versions = set(policy["symbol_versions"][arch][prefix])
                if not symbol_versions_arch[prefix].issubset(policy_symbol_versions):
                    diff = symbol_versions_arch[prefix] - policy_symbol_versions
                    raise ValueError(
                        'Invalid "policy.json" file. Symbol versions missing '
                        f'in "{policy["name"]}_{arch}" for "{prefix}" '
                        f"compared to previous policies: {diff}"
                    )
                symbol_versions_arch[prefix].update(
                    policy["symbol_versions"][arch][prefix]
                )
            symbol_versions[arch] = symbol_versions_arch


_POLICY_JSON_MAP = {
    Libc.GLIBC: _HERE / "manylinux-policy.json",
    Libc.MUSL: _HERE / "musllinux-policy.json",
}


def _get_musl_policy():
    if _LIBC != Libc.MUSL:
        return None
    musl_version = get_musl_version(find_musl_libc())
    return f"musllinux_{musl_version.major}_{musl_version.minor}"


_MUSL_POLICY = _get_musl_policy()


def _fixup_musl_libc_soname(whitelist):
    if _LIBC != Libc.MUSL:
        return whitelist
    soname_map = {
        "libc.so": {
            "x86_64": "libc.musl-x86_64.so.1",
            "i686": "libc.musl-x86.so.1",
            "aarch64": "libc.musl-aarch64.so.1",
            "s390x": "libc.musl-s390x.so.1",
            "ppc64le": "libc.musl-ppc64le.so.1",
            "armv7l": "libc.musl-armv7.so.1",
        }
    }
    new_whitelist = []
    for soname in whitelist:
        if soname in soname_map:
            new_soname = soname_map[soname][_ARCH_NAME]
            logger.debug(f"Replacing whitelisted '{soname}' by '{new_soname}'")
            new_whitelist.append(new_soname)
        else:
            new_whitelist.append(soname)
    return new_whitelist


with _POLICY_JSON_MAP[_LIBC].open() as f:
    _POLICIES = []
    _policies_temp = json.load(f)
    _validate_pep600_compliance(_policies_temp)
    for _p in _policies_temp:
        if _MUSL_POLICY is not None and _p["name"] not in {"linux", _MUSL_POLICY}:
            continue
        if _ARCH_NAME in _p["symbol_versions"].keys() or _p["name"] == "linux":
            if _p["name"] != "linux":
                _p["symbol_versions"] = _p["symbol_versions"][_ARCH_NAME]
            _p["name"] = _p["name"] + "_" + _ARCH_NAME
            _p["aliases"] = [alias + "_" + _ARCH_NAME for alias in _p["aliases"]]
            _p["lib_whitelist"] = _fixup_musl_libc_soname(_p["lib_whitelist"])
            _POLICIES.append(_p)
    if _LIBC == Libc.MUSL:
        assert len(_POLICIES) == 2, _POLICIES

POLICY_PRIORITY_HIGHEST = max(p["priority"] for p in _POLICIES)
POLICY_PRIORITY_LOWEST = min(p["priority"] for p in _POLICIES)


def load_policies():
    return _POLICIES


def _load_policy_schema():
    with open(join(dirname(abspath(__file__)), "policy-schema.json")) as f_:
        schema = json.load(f_)
    return schema


def get_policy_by_name(name: str) -> Optional[Dict]:
    matches = [p for p in _POLICIES if p["name"] == name or name in p["aliases"]]
    if len(matches) == 0:
        return None
    if len(matches) > 1:
        raise RuntimeError("Internal error. Policies should be unique")
    return matches[0]


def get_policy_name(priority: int) -> Optional[str]:
    matches = [p["name"] for p in _POLICIES if p["priority"] == priority]
    if len(matches) == 0:
        return None
    if len(matches) > 1:
        raise RuntimeError("Internal error. priorities should be unique")
    return matches[0]


def get_priority_by_name(name: str) -> Optional[int]:
    policy = get_policy_by_name(name)
    return None if policy is None else policy["priority"]


def get_replace_platforms(name: str) -> List[str]:
    """Extract platform tag replacement rules from policy

    >>> get_replace_platforms('linux_x86_64')
    []
    >>> get_replace_platforms('linux_i686')
    []
    >>> get_replace_platforms('manylinux1_x86_64')
    ['linux_x86_64']
    >>> get_replace_platforms('manylinux1_i686')
    ['linux_i686']

    """
    if name.startswith("linux"):
        return []
    if name.startswith("manylinux_"):
        return ["linux_" + "_".join(name.split("_")[3:])]
    if name.startswith("musllinux_"):
        return ["linux_" + "_".join(name.split("_")[3:])]
    return ["linux_" + "_".join(name.split("_")[1:])]


# These have to be imported here to avoid a circular import.
from .external_references import lddtree_external_references  # noqa
from .versioned_symbols import versioned_symbols_policy  # noqa

__all__ = [
    "lddtree_external_references",
    "versioned_symbols_policy",
    "load_policies",
    "POLICY_PRIORITY_HIGHEST",
    "POLICY_PRIORITY_LOWEST",
]



import os

import setuptools.command.build_ext
from setuptools import Distribution, find_packages, setup
from setuptools.extension import Extension, Library

# despite its name, setuptools.command.build_ext.link_shared_object won't
# link a shared object on Linux, but a static library and patches distutils
# for this ... We're patching this back now.


def always_link_shared_object(
    self,
    objects,
    output_libname,
    output_dir=None,
    libraries=None,
    library_dirs=None,
    runtime_library_dirs=None,
    export_symbols=None,
    debug=0,
    extra_preargs=None,
    extra_postargs=None,
    build_temp=None,
    target_lang=None,
):
    self.link(
        self.SHARED_LIBRARY,
        objects,
        output_libname,
        output_dir,
        libraries,
        library_dirs,
        runtime_library_dirs,
        export_symbols,
        debug,
        extra_preargs,
        extra_postargs,
        build_temp,
        target_lang,
    )


setuptools.command.build_ext.libtype = "shared"
setuptools.command.build_ext.link_shared_object = always_link_shared_object

libtype = setuptools.command.build_ext.libtype
build_ext_cmd = Distribution().get_command_obj("build_ext")
build_ext_cmd.initialize_options()
build_ext_cmd.setup_shlib_compiler()


def libname(name):
    """gets 'name' and returns something like libname.cpython-37m-darwin.so"""
    filename = build_ext_cmd.get_ext_filename(name)
    fn, ext = os.path.splitext(filename)
    return build_ext_cmd.shlib_compiler.library_filename(fn, libtype)


pkg_name = "nonpy_rpath"
crypt_name = "_cryptexample"
crypt_soname = libname(crypt_name)

build_cmd = Distribution().get_command_obj("build")
build_cmd.finalize_options()
build_platlib = build_cmd.build_platlib


def link_args(soname=None):
    args = []
    if soname:
        args += ["-Wl,-soname," + soname]
    loader_path = "$ORIGIN"
    args += ["-Wl,-rpath," + loader_path]
    return args


nonpy_rpath_module = Extension(
    pkg_name + "._nonpy_rpath",
    language="c++",
    sources=["nonpy_rpath.cpp"],
    extra_link_args=link_args(),
    extra_objects=[build_platlib + "/nonpy_rpath/" + crypt_soname],
)
crypt_example = Library(
    pkg_name + "." + crypt_name,
    language="c++",
    extra_compile_args=["-lcrypt"],
    extra_link_args=link_args(crypt_soname) + ["-lcrypt"],
    sources=["extensions/testcrypt.cpp"],
)

setup(
    name="nonpy_rpath",
    version="0.1.0",
    packages=find_packages(),
    description="Test package for nonpy_rpath",
    ext_modules=[crypt_example, nonpy_rpath_module],
)



# Copyright 2016 Robert McGibbon
# Copyright 2012-2014 Gentoo Foundation
# Copyright 2012-2014 Mike Frysinger <vapier@gentoo.org>
# Copyright 2012-2014 The Chromium OS Authors
# Use of this source code is governed by a BSD-style license (BSD-3)
# Original version available from:
#   https://sources.gentoo.org/cgi-bin/viewvc.cgi/gentoo-projects/pax-utils/lddtree.py
"""Read the ELF dependency tree

This does not work like `ldd` in that we do not execute/load code (only read
files on disk), and we parse the dependency structure as a tree rather than
 a flat list.
"""

import errno
import functools
import glob
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from elftools.elf.elffile import ELFFile

from .libc import Libc, get_libc

log = logging.getLogger(__name__)
__all__ = ["lddtree"]


def normpath(path: str) -> str:
    """Normalize a path

    Python's os.path.normpath() doesn't handle some cases:
    // -> //
    //..// -> //
    //..//..// -> ///
    """
    return os.path.normpath(path).replace("//", "/")


def readlink(path: str, root: str, prefixed: bool = False) -> str:
    """Like os.readlink(), but relative to a ``root``

    This does not currently handle the pathological case:
    /lib/foo.so -> ../../../../../../../foo.so
    This relies on the .. entries in / to point to itself.

    Parameters
    ----------
    path
        The symlink to read
    root
        The path to use for resolving absolute symlinks
    prefixed
        When False, the ``path`` must not have ``root`` prefixed to it, nor
        will the return value have ``root`` prefixed.  When True, ``path``
        must have ``root`` prefixed, and the return value will have ``root``
        added.

    Returns
    -------
    A fully resolved symlink path
    """
    root = root.rstrip("/")
    if prefixed:
        path = path[len(root) :]

    while os.path.islink(root + path):
        path = os.path.join(os.path.dirname(path), os.readlink(root + path))

    return normpath((root + path) if prefixed else path)


def dedupe(items: List[str]) -> List[str]:
    """Remove all duplicates from ``items`` (keeping order)"""
    seen = {}  # type: Dict[str, str]
    return [seen.setdefault(x, x) for x in items if x not in seen]


def parse_ld_paths(str_ldpaths: str, path: str, root: str = "") -> List[str]:
    """Parse the colon-delimited list of paths and apply ldso rules to each

    Note the special handling as dictated by the ldso:
    - Empty paths are equivalent to $PWD
    - $ORIGIN is expanded to the path of the given file
    - (TODO) $LIB and friends

    Parameters
    ----------
    str_ldpaths
        A colon-delimited string of paths
    root
        The path to prepend to all paths found
    path
        The object actively being parsed (used for $ORIGIN)

    Returns
    -------
        list of processed paths
    """
    ldpaths = []  # type: List[str]
    for ldpath in str_ldpaths.split(":"):
        if ldpath == "":
            # The ldso treats "" paths as $PWD.
            ldpath = os.getcwd()
        elif "$ORIGIN" in ldpath:
            ldpath = ldpath.replace("$ORIGIN", os.path.dirname(os.path.abspath(path)))
        else:
            ldpath = root + ldpath
        ldpaths.append(normpath(ldpath))
    return [p for p in dedupe(ldpaths) if os.path.isdir(p)]


@functools.lru_cache()
def parse_ld_so_conf(ldso_conf: str, root: str = "/", _first: bool = True) -> List[str]:
    """Load all the paths from a given ldso config file

    This should handle comments, whitespace, and "include" statements.

    Parameters
    ----------
    ldso_conf
        The file to scan
    root
        The path to prepend to all paths found
    _first
        Recursive use only; is this the first ELF?

    Returns
    -------
    list of paths found
    """
    paths = []  # type: List[str]

    dbg_pfx = "" if _first else "  "
    try:
        log.debug("%sparse_ld_so_conf(%s)", dbg_pfx, ldso_conf)
        with open(ldso_conf) as f:
            for line in f.readlines():
                line = line.split("#", 1)[0].strip()
                if not line:
                    continue
                if line.startswith("include "):
                    line = line[8:]
                    if line[0] == "/":
                        line = root + line.lstrip("/")
                    else:
                        line = os.path.dirname(ldso_conf) + "/" + line
                    log.debug("%s  glob: %s", dbg_pfx, line)
                    for path in glob.glob(line):
                        paths += parse_ld_so_conf(path, root=root, _first=False)
                else:
                    paths += [normpath(root + line)]
    except OSError as e:
        if e.errno != errno.ENOENT:
            log.warning(e)

    if _first:
        # XXX: Load paths from ldso itself.
        # Remove duplicate entries to speed things up.
        paths = [p for p in dedupe(paths) if os.path.isdir(p)]

    return paths


@functools.lru_cache()
def load_ld_paths(root: str = "/", prefix: str = "") -> Dict[str, List[str]]:
    """Load linker paths from common locations

    This parses the ld.so.conf and LD_LIBRARY_PATH env var.

    Parameters
    ----------
    root
        The root tree to prepend to paths
    prefix
        The path under ``root`` to search

    Returns
    -------
    dict containing library paths to search
    """
    ldpaths = {"conf": [], "env": [], "interp": []}  # type: Dict

    # Load up $LD_LIBRARY_PATH.
    env_ldpath = os.environ.get("LD_LIBRARY_PATH")
    if env_ldpath is not None:
        if root != "/":
            log.warning("ignoring LD_LIBRARY_PATH due to ROOT usage")
        else:
            # XXX: If this contains $ORIGIN, we probably have to parse this
            # on a per-ELF basis so it can get turned into the right thing.
            ldpaths["env"] = parse_ld_paths(env_ldpath, path="")

    libc = get_libc()
    if libc == Libc.MUSL:
        # from https://git.musl-libc.org/cgit/musl/tree/ldso
        # /dynlink.c?id=3f701faace7addc75d16dea8a6cd769fa5b3f260#n1063
        root_prefix = Path(root) / prefix
        ld_musl = list((root_prefix / "etc").glob("ld-musl-*.path"))
        assert len(ld_musl) <= 1
        if len(ld_musl) == 0:
            ldpaths["conf"] = [
                root + "/lib",
                root + "/usr/local/lib",
                root + "/usr/lib",
            ]
        else:
            ldpaths["conf"] = []
            for ldpath in ld_musl[0].read_text().split(":"):
                ldpath_stripped = ldpath.strip()
                if ldpath_stripped == "":
                    continue
                ldpaths["conf"].append(root + ldpath_stripped)
    else:
        # Load up /etc/ld.so.conf.
        ldpaths["conf"] = parse_ld_so_conf(root + prefix + "/etc/ld.so.conf", root=root)
        # the trusted directories are not necessarily in ld.so.conf
        ldpaths["conf"].extend(["/lib", "/lib64/", "/usr/lib", "/usr/lib64"])
    log.debug("linker ldpaths: %s", ldpaths)
    return ldpaths


def compatible_elfs(elf1: ELFFile, elf2: ELFFile) -> bool:
    """See if two ELFs are compatible

    This compares the aspects of the ELF to see if they're compatible:
    bit size, endianness, machine type, and operating system.

    Parameters
    ----------
    elf1 : ELFFile
    elf2 : ELFFile

    Returns
    -------
    True if compatible, False otherwise
    """
    osabis = frozenset(e.header["e_ident"]["EI_OSABI"] for e in (elf1, elf2))
    compat_sets = (
        frozenset(
            "ELFOSABI_%s" % x
            for x in (
                "NONE",
                "SYSV",
                "GNU",
                "LINUX",
            )
        ),
    )
    return (
        (len(osabis) == 1 or any(osabis.issubset(x) for x in compat_sets))
        and elf1.elfclass == elf2.elfclass
        and elf1.little_endian == elf2.little_endian
        and elf1.header["e_machine"] == elf2.header["e_machine"]
    )


def find_lib(
    elf: ELFFile, lib: str, ldpaths: List[str], root: str = "/"
) -> Tuple[Optional[str], Optional[str]]:
    """Try to locate a ``lib`` that is compatible to ``elf`` in the given
    ``ldpaths``

    Parameters
    ----------
    elf : ELFFile
        The elf which the library should be compatible with (ELF wise)
    lib : str
        The library (basename) to search for
    ldpaths : List[str]
        A list of paths to search
    root : str
       The root path to resolve symlinks

    Returns
    -------
    Tuple of the full path to the desired library and the real path to it
    """

    for ldpath in ldpaths:
        path = os.path.join(ldpath, lib)
        target = readlink(path, root, prefixed=True)

        if os.path.exists(target):
            with open(target, "rb") as f:
                libelf = ELFFile(f)
                if compatible_elfs(elf, libelf):
                    return (target, path)

    return (None, None)


def lddtree(
    path: str,
    root: str = "/",
    prefix: str = "",
    ldpaths: Optional[Dict[str, List[str]]] = None,
    display: Optional[str] = None,
    _first: bool = True,
    _all_libs: Dict = {},
) -> Dict:
    """Parse the ELF dependency tree of the specified file

    Parameters
    ----------
    path
        The ELF to scan
    root
        The root tree to prepend to paths; this applies to interp and rpaths
        only as ``path`` and ``ldpaths`` are expected to be prefixed already
    prefix
        The path under ``root`` to search
    ldpaths
        dict containing library paths to search; should have the keys:
        conf, env, interp. If not supplied, the function ``load_ld_paths``
        will be called.
    display
        The path to show rather than ``path``
    _first
        Recursive use only; is this the first ELF?
    _all_libs
        Recursive use only; dict of all libs we've seen

    Returns
    -------
    a dict containing information about all the ELFs; e.g.
    {
      'interp': '/lib64/ld-linux.so.2',
      'needed': ['libc.so.6', 'libcurl.so.4',],
      'libs': {
        'libc.so.6': {
          'path': '/lib64/libc.so.6',
          'needed': [],
        },
        'libcurl.so.4': {
          'path': '/usr/lib64/libcurl.so.4',
          'needed': ['libc.so.6', 'librt.so.1',],
        },
      },
    }
    """
    if not ldpaths:
        ldpaths = load_ld_paths().copy()

    if _first:
        _all_libs = {}

    ret = {
        "interp": None,
        "path": path if display is None else display,
        "realpath": path,
        "needed": [],
        "rpath": [],
        "runpath": [],
        "libs": _all_libs,
    }  # type: Dict[str, Any]

    log.debug("lddtree(%s)" % path)

    with open(path, "rb") as f:
        elf = ELFFile(f)

        # If this is the first ELF, extract the interpreter.
        if _first:
            for segment in elf.iter_segments():
                if segment.header.p_type != "PT_INTERP":
                    continue

                interp = segment.get_interp_name()
                log.debug("  interp           = %s", interp)
                ret["interp"] = normpath(root + interp)
                ret["libs"][os.path.basename(interp)] = {
                    "path": ret["interp"],
                    "realpath": readlink(ret["interp"], root, prefixed=True),
                    "needed": [],
                }
                # XXX: Should read it and scan for /lib paths.
                ldpaths["interp"] = [
                    normpath(root + os.path.dirname(interp)),
                    normpath(
                        root + prefix + "/usr" + os.path.dirname(interp).lstrip(prefix)
                    ),
                ]
                log.debug("  ldpaths[interp]  = %s", ldpaths["interp"])
                break

        # Parse the ELF's dynamic tags.
        libs = []  # type: List[str]
        rpaths = []  # type: List[str]
        runpaths = []  # type: List[str]
        for segment in elf.iter_segments():
            if segment.header.p_type != "PT_DYNAMIC":
                continue

            for t in segment.iter_tags():
                if t.entry.d_tag == "DT_RPATH":
                    rpaths = parse_ld_paths(t.rpath, path=path, root=root)
                elif t.entry.d_tag == "DT_RUNPATH":
                    runpaths = parse_ld_paths(t.runpath, path=path, root=root)
                elif t.entry.d_tag == "DT_NEEDED":
                    libs.append(t.needed)
            if runpaths:
                # If both RPATH and RUNPATH are set, only the latter is used.
                rpaths = []

            # XXX: We assume there is only one PT_DYNAMIC.  This is
            # probably fine since the runtime ldso does the same.
            break
        if _first:
            # Propagate the rpaths used by the main ELF since those will be
            # used at runtime to locate things.
            ldpaths["rpath"] = rpaths
            ldpaths["runpath"] = runpaths
            log.debug("  ldpaths[rpath]   = %s", rpaths)
            log.debug("  ldpaths[runpath] = %s", runpaths)
        ret["rpath"] = rpaths
        ret["runpath"] = runpaths
        ret["needed"] = libs

        # Search for the libs this ELF uses.
        all_ldpaths = None  # type: Optional[List[str]]
        for lib in libs:
            if lib in _all_libs:
                continue
            if all_ldpaths is None:
                all_ldpaths = (
                    ldpaths["rpath"]
                    + rpaths
                    + runpaths
                    + ldpaths["env"]
                    + ldpaths["runpath"]
                    + ldpaths["conf"]
                    + ldpaths["interp"]
                )
            realpath, fullpath = find_lib(elf, lib, all_ldpaths, root)
            _all_libs[lib] = {
                "realpath": realpath,
                "path": fullpath,
                "needed": [],
            }
            if realpath and fullpath:
                lret = lddtree(
                    realpath,
                    root,
                    prefix,
                    ldpaths,
                    display=fullpath,
                    _first=False,
                    _all_libs=_all_libs,
                )
                _all_libs[lib]["needed"] = lret["needed"]

        del elf

    return ret



""" Contexts for *with* statement providing temporary directories
"""
import os
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Optional, Type


class InTemporaryDirectory:
    """Create, return, and change directory to a temporary directory

    Examples
    --------
    >>> import os
    >>> my_cwd = os.getcwd()
    >>> with InTemporaryDirectory() as tmpdir:
    ...     _ = open('test.txt', 'wt').write('some text')
    ...     assert os.path.isfile('test.txt')
    ...     assert os.path.isfile(os.path.join(tmpdir, 'test.txt'))
    >>> os.path.exists(tmpdir)
    False
    >>> os.getcwd() == my_cwd
    True
    """

    def __init__(self) -> None:
        self._tmpdir = TemporaryDirectory()

    @property
    def name(self) -> str:
        return self._tmpdir.name

    def __enter__(self) -> str:
        self._pwd = os.getcwd()
        os.chdir(self._tmpdir.name)
        return self._tmpdir.__enter__()

    def __exit__(
        self,
        exc: Optional[Type[BaseException]],
        value: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        os.chdir(self._pwd)
        return self._tmpdir.__exit__(exc, value, tb)


class InGivenDirectory:
    """Change directory to given directory for duration of ``with`` block

    Useful when you want to use `InTemporaryDirectory` for the final test, but
    you are still debugging.  For example, you may want to do this in the end:

    >>> with InTemporaryDirectory() as tmpdir:
    ...     # do something complicated which might break
    ...     pass

    But indeed the complicated thing does break, and meanwhile the
    ``InTemporaryDirectory`` context manager wiped out the directory with the
    temporary files that you wanted for debugging.  So, while debugging, you
    replace with something like:

    >>> with InGivenDirectory() as tmpdir: # Use working directory by default
    ...     # do something complicated which might break
    ...     pass

    You can then look at the temporary file outputs to debug what is happening,
    fix, and finally replace ``InGivenDirectory`` with ``InTemporaryDirectory``
    again.
    """

    def __init__(self, path: Optional[str] = None) -> None:
        """Initialize directory context manager

        Parameters
        ----------
        path : None or str, optional
            path to change directory to, for duration of ``with`` block.
            Defaults to ``os.getcwd()`` if None
        """
        if path is None:
            path = os.getcwd()
        self.name = os.path.abspath(path)

    def __enter__(self) -> str:
        self._pwd = os.path.abspath(os.getcwd())
        if not os.path.isdir(self.name):
            os.mkdir(self.name)
        os.chdir(self.name)
        return self.name

    def __exit__(
        self,
        exc: Optional[Type[BaseException]],
        value: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        os.chdir(self._pwd)



import subprocess

import pkg_resources


def runit(x):
    filename = pkg_resources.resource_filename(__name__, "testprogram")
    output = subprocess.check_output([filename, str(x)])
    return float(output)



"""Context managers like those in wheeltools.py for unpacking
conda packages.
"""
import os
from typing import List, Optional

from .tmpdirs import InTemporaryDirectory
from .tools import tarbz2todir


class InCondaPkg(InTemporaryDirectory):
    def __init__(self, in_conda_pkg: str) -> None:
        """Initialize in-conda-package context manager"""
        self.in_conda_pkg = os.path.abspath(in_conda_pkg)
        super().__init__()

    def __enter__(self) -> str:
        tarbz2todir(self.in_conda_pkg, self.name)
        return super().__enter__()


class InCondaPkgCtx(InCondaPkg):
    def __init__(self, in_conda_pkg: str) -> None:
        super().__init__(in_conda_pkg)
        self.path: Optional[str] = None

    def __enter__(self):
        self.path = super().__enter__()
        return self

    def iter_files(self) -> List[str]:
        if self.path is None:
            raise ValueError("This function should be called from context manager")
        files = os.path.join(self.path, "info", "files")
        with open(files) as f:
            return [line.strip() for line in f.readlines()]



import logging

logger = logging.getLogger(__name__)


def configure_subparser(sub_parsers):
    help = "Analyze a single ELF file (similar to ``ldd``)."
    p = sub_parsers.add_parser("lddtree", help=help, description=help)
    p.add_argument("file", help="Path to .so file")
    p.set_defaults(func=execute)


def execute(args, p):
    import json

    from .lddtree import lddtree

    logger.info(json.dumps(lddtree(args.file), indent=4))



import subprocess
from os import getenv, path

from setuptools import Extension, setup

cmd = "gcc -shared -fPIC -D_GNU_SOURCE dependency.c -o libdependency.so -lm -lc"
subprocess.check_call(cmd.split())

define_macros = [("_GNU_SOURCE", None)]
libraries = []
library_dirs = []

if getenv("WITH_DEPENDENCY", "0") == "1":
    libraries.append("dependency")
    library_dirs.append(path.abspath(path.dirname(__file__)))
    define_macros.append(("WITH_DEPENDENCY", "1"))

libraries.extend(["m", "c"])

setup(
    name="testdependencies",
    version="0.0.1",
    ext_modules=[
        Extension(
            "testdependencies",
            sources=["testdependencies.c"],
            define_macros=define_macros,
            libraries=libraries,
            library_dirs=library_dirs,
        )
    ],
)



import argparse
import logging
from os.path import abspath, basename, exists, isfile

from auditwheel.patcher import Patchelf

from .policy import (
    POLICY_PRIORITY_HIGHEST,
    get_policy_by_name,
    get_policy_name,
    get_priority_by_name,
    load_policies,
)
from .tools import EnvironmentDefault

logger = logging.getLogger(__name__)


def configure_parser(sub_parsers):
    policies = load_policies()
    policy_names = [p["name"] for p in policies]
    policy_names += [alias for p in policies for alias in p["aliases"]]
    epilog = """PLATFORMS:
These are the possible target platform tags, as specified by PEP 600.
Note that old, pre-PEP 600 tags are still usable and are listed as aliases
below.
"""
    for p in policies:
        epilog += f"- {p['name']}"
        if len(p["aliases"]) > 0:
            epilog += f" (aliased by {', '.join(p['aliases'])})"
        epilog += "\n"
    highest_policy = get_policy_name(POLICY_PRIORITY_HIGHEST)
    help = """Vendor in external shared library dependencies of a wheel.
If multiple wheels are specified, an error processing one
wheel will abort processing of subsequent wheels.
"""
    p = sub_parsers.add_parser(
        "repair",
        help=help,
        description=help,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    p.add_argument("WHEEL_FILE", help="Path to wheel file.", nargs="+")
    p.add_argument(
        "--plat",
        action=EnvironmentDefault,
        metavar="PLATFORM",
        env="AUDITWHEEL_PLAT",
        dest="PLAT",
        help="Desired target platform. See the available platforms under the "
        f'PLATFORMS section below. (default: "{highest_policy}")',
        choices=policy_names,
        default=highest_policy,
    )
    p.add_argument(
        "-L",
        "--lib-sdir",
        dest="LIB_SDIR",
        help=(
            "Subdirectory in packages to store copied libraries." ' (default: ".libs")'
        ),
        default=".libs",
    )
    p.add_argument(
        "-w",
        "--wheel-dir",
        dest="WHEEL_DIR",
        type=abspath,
        help=("Directory to store delocated wheels (default:" ' "wheelhouse/")'),
        default="wheelhouse/",
    )
    p.add_argument(
        "--no-update-tags",
        dest="UPDATE_TAGS",
        action="store_false",
        help=(
            "Do not update the wheel filename tags and WHEEL info"
            " to match the repaired platform tag."
        ),
        default=True,
    )
    p.add_argument(
        "--strip",
        dest="STRIP",
        action="store_true",
        help="Strip symbols in the resulting wheel",
        default=False,
    )
    p.add_argument(
        "--exclude",
        dest="EXCLUDE",
        help="Exclude SONAME from grafting into the resulting wheel "
        "(can be specified multiple times)",
        action="append",
        default=[],
    )
    p.add_argument(
        "--only-plat",
        dest="ONLY_PLAT",
        action="store_true",
        help="Do not check for higher policy compatibility",
        default=False,
    )
    p.set_defaults(func=execute)


def execute(args, p):
    import os

    from .repair import repair_wheel
    from .wheel_abi import NonPlatformWheel, analyze_wheel_abi

    for wheel_file in args.WHEEL_FILE:
        if not isfile(wheel_file):
            p.error("cannot access %s. No such file" % wheel_file)

        logger.info("Repairing %s", basename(wheel_file))

        if not exists(args.WHEEL_DIR):
            os.makedirs(args.WHEEL_DIR)

        try:
            wheel_abi = analyze_wheel_abi(wheel_file)
        except NonPlatformWheel:
            logger.info(NonPlatformWheel.LOG_MESSAGE)
            return 1

        policy = get_policy_by_name(args.PLAT)
        reqd_tag = policy["priority"]

        if reqd_tag > get_priority_by_name(wheel_abi.sym_tag):
            msg = (
                'cannot repair "%s" to "%s" ABI because of the presence '
                "of too-recent versioned symbols. You'll need to compile "
                "the wheel on an older toolchain." % (wheel_file, args.PLAT)
            )
            p.error(msg)

        if reqd_tag > get_priority_by_name(wheel_abi.ucs_tag):
            msg = (
                'cannot repair "%s" to "%s" ABI because it was compiled '
                "against a UCS2 build of Python. You'll need to compile "
                "the wheel against a wide-unicode build of Python."
                % (wheel_file, args.PLAT)
            )
            p.error(msg)

        if reqd_tag > get_priority_by_name(wheel_abi.blacklist_tag):
            msg = (
                'cannot repair "%s" to "%s" ABI because it depends on '
                "black-listed symbols." % (wheel_file, args.PLAT)
            )
            p.error(msg)

        abis = [policy["name"]] + policy["aliases"]
        if not args.ONLY_PLAT:
            if reqd_tag < get_priority_by_name(wheel_abi.overall_tag):
                logger.info(
                    (
                        "Wheel is eligible for a higher priority tag. "
                        "You requested %s but I have found this wheel is "
                        "eligible for %s."
                    ),
                    args.PLAT,
                    wheel_abi.overall_tag,
                )
                higher_policy = get_policy_by_name(wheel_abi.overall_tag)
                abis = [higher_policy["name"]] + higher_policy["aliases"] + abis

        patcher = Patchelf()
        out_wheel = repair_wheel(
            wheel_file,
            abis=abis,
            lib_sdir=args.LIB_SDIR,
            out_dir=args.WHEEL_DIR,
            update_tags=args.UPDATE_TAGS,
            patcher=patcher,
            exclude=args.EXCLUDE,
            strip=args.STRIP,
        )

        if out_wheel is not None:
            logger.info("\nFixed-up wheel written to %s", out_wheel)



from unittest.mock import patch

import pytest

from auditwheel.policy import (
    _validate_pep600_compliance,
    get_arch_name,
    get_policy_name,
    get_priority_by_name,
    get_replace_platforms,
)


@patch("auditwheel.policy._platform_module.machine")
@patch("auditwheel.policy.bits", 32)
@pytest.mark.parametrize(
    "reported_arch,expected_arch",
    [
        ("armv6l", "armv6l"),
        ("armv7l", "armv7l"),
        ("i686", "i686"),
        ("x86_64", "i686"),
    ],
)
def test_32bits_arch_name(machine_mock, reported_arch, expected_arch):
    machine_mock.return_value = reported_arch
    machine = get_arch_name()
    assert machine == expected_arch


@patch("auditwheel.policy._platform_module.machine")
@patch("auditwheel.policy.bits", 64)
@pytest.mark.parametrize(
    "reported_arch,expected_arch",
    [
        ("aarch64", "aarch64"),
        ("ppc64le", "ppc64le"),
        ("x86_64", "x86_64"),
    ],
)
def test_64bits_arch_name(machine_mock, reported_arch, expected_arch):
    machine_mock.return_value = reported_arch
    machine = get_arch_name()
    assert machine == expected_arch


@pytest.mark.parametrize(
    "name,expected",
    [
        ("linux_aarch64", []),
        ("manylinux1_ppc64le", ["linux_ppc64le"]),
        ("manylinux2014_x86_64", ["linux_x86_64"]),
        ("manylinux_2_24_x86_64", ["linux_x86_64"]),
    ],
)
def test_replacement_platform(name, expected):
    assert get_replace_platforms(name) == expected


def test_pep600_compliance():
    _validate_pep600_compliance(
        [
            {
                "name": "manylinux1",
                "priority": 100,
                "symbol_versions": {
                    "i686": {"CXXABI": ["1.3"]},
                },
                "lib_whitelist": ["libgcc_s.so.1"],
            },
            {
                "name": "manylinux2010",
                "priority": 90,
                "symbol_versions": {
                    "i686": {"CXXABI": ["1.3", "1.3.1"]},
                },
                "lib_whitelist": ["libgcc_s.so.1", "libstdc++.so.6"],
            },
        ]
    )

    _validate_pep600_compliance(
        [
            {
                "name": "manylinux1",
                "priority": 100,
                "symbol_versions": {
                    "i686": {"CXXABI": ["1.3"]},
                    "x86_64": {"CXXABI": ["1.3"]},
                },
                "lib_whitelist": ["libgcc_s.so.1"],
            },
            {
                "name": "manylinux2010",
                "priority": 90,
                "symbol_versions": {
                    "i686": {"CXXABI": ["1.3", "1.3.1"]},
                },
                "lib_whitelist": ["libgcc_s.so.1", "libstdc++.so.6"],
            },
        ]
    )

    with pytest.raises(ValueError, match=r"manylinux2010_i686.*CXXABI.*1.3.2"):
        _validate_pep600_compliance(
            [
                {
                    "name": "manylinux1",
                    "priority": 100,
                    "symbol_versions": {
                        "i686": {"CXXABI": ["1.3", "1.3.2"]},
                    },
                    "lib_whitelist": ["libgcc_s.so.1"],
                },
                {
                    "name": "manylinux2010",
                    "priority": 90,
                    "symbol_versions": {
                        "i686": {"CXXABI": ["1.3", "1.3.1"]},
                    },
                    "lib_whitelist": ["libgcc_s.so.1", "libstdc++.so.6"],
                },
            ]
        )

    with pytest.raises(ValueError, match=r"manylinux2010.*libstdc\+\+\.so\.6"):
        _validate_pep600_compliance(
            [
                {
                    "name": "manylinux1",
                    "priority": 100,
                    "symbol_versions": {
                        "i686": {"CXXABI": ["1.3"]},
                    },
                    "lib_whitelist": ["libgcc_s.so.1", "libstdc++.so.6"],
                },
                {
                    "name": "manylinux2010",
                    "priority": 90,
                    "symbol_versions": {
                        "i686": {"CXXABI": ["1.3", "1.3.1"]},
                    },
                    "lib_whitelist": ["libgcc_s.so.1"],
                },
            ]
        )


class TestPolicyAccess:
    def test_get_by_priority(self):
        _arch = get_arch_name()
        assert get_policy_name(65) == f"manylinux_2_27_{_arch}"
        assert get_policy_name(70) == f"manylinux_2_24_{_arch}"
        assert get_policy_name(80) == f"manylinux_2_17_{_arch}"
        if _arch in {"x86_64", "i686"}:
            assert get_policy_name(90) == f"manylinux_2_12_{_arch}"
            assert get_policy_name(100) == f"manylinux_2_5_{_arch}"
        assert get_policy_name(0) == f"linux_{_arch}"

    def test_get_by_priority_missing(self):
        assert get_policy_name(101) is None

    @patch(
        "auditwheel.policy._POLICIES",
        [
            {"name": "duplicate", "priority": 0},
            {"name": "duplicate", "priority": 0},
        ],
    )
    def test_get_by_priority_duplicate(self):
        with pytest.raises(RuntimeError):
            get_policy_name(0)

    def test_get_by_name(self):
        _arch = get_arch_name()
        assert get_priority_by_name(f"manylinux_2_27_{_arch}") == 65
        assert get_priority_by_name(f"manylinux_2_24_{_arch}") == 70
        assert get_priority_by_name(f"manylinux2014_{_arch}") == 80
        assert get_priority_by_name(f"manylinux_2_17_{_arch}") == 80
        if _arch in {"x86_64", "i686"}:
            assert get_priority_by_name(f"manylinux2010_{_arch}") == 90
            assert get_priority_by_name(f"manylinux_2_12_{_arch}") == 90
            assert get_priority_by_name(f"manylinux1_{_arch}") == 100
            assert get_priority_by_name(f"manylinux_2_5_{_arch}") == 100

    def test_get_by_name_missing(self):
        assert get_priority_by_name("nosuchpolicy") is None

    @patch(
        "auditwheel.policy._POLICIES",
        [
            {"name": "duplicate", "priority": 0},
            {"name": "duplicate", "priority": 0},
        ],
    )
    def test_get_by_name_duplicate(self):
        with pytest.raises(RuntimeError):
            get_priority_by_name("duplicate")



from setuptools import Extension, find_packages, setup

setup(
    name="multiple_top_level",
    version="1.0",
    description="Auditwheel multiple top-level extensions example",
    packages=find_packages(where="src"),
    ext_modules=[
        Extension(
            "example_a",
            ["src/example_a.pyx"],
            include_dirs=["lib-src/a"],
            library_dirs=["lib-src/a", "lib-src/b"],
            libraries=["a"],
        ),
        Extension(
            "example_b",
            ["src/example_b.pyx"],
            include_dirs=["lib-src/a"],
            library_dirs=["lib-src/a", "lib-src/b"],
            libraries=["a"],
        ),
    ],
)



from pathlib import Path
from unittest.mock import Mock

from auditwheel.main_addtag import execute

HERE = Path(__file__).parent.resolve()


def test_smoke(tmpdir):
    """Simple test to exercise the 'addtag' code path"""
    args = Mock(
        WHEEL_FILE=str(HERE / "cffi-1.5.0-cp27-none-linux_x86_64.whl"),
        WHEEL_DIR=str(tmpdir / "wheelhouse/"),
    )
    execute(args, None)



import os

import pretend
import pytest

from auditwheel import wheel_abi


class TestGetWheelElfdata:
    @pytest.mark.parametrize(
        "filenames, message",
        [
            (
                # A single invalid file
                [os.sep.join(["purelib", "foo"])],
                "Invalid binary wheel, found the following shared library/libraries in"
                " purelib folder:\n\tfoo\nThe wheel has to be platlib compliant in "
                "order to be repaired by auditwheel.",
            ),
            (
                # Multiple invalid files
                [os.sep.join(["purelib", "foo"]), os.sep.join(["purelib", "bar"])],
                "Invalid binary wheel, found the following shared library/libraries in"
                " purelib folder:\n\tfoo\n\tbar\nThe wheel has to be platlib compliant"
                " in order to be repaired by auditwheel.",
            ),
        ],
    )
    def test_finds_shared_library_in_purelib(self, filenames, message, monkeypatch):
        entered_context = pretend.stub(iter_files=lambda: filenames)
        context = pretend.stub(
            __enter__=lambda: entered_context, __exit__=lambda *a: None
        )
        InGenericPkgCtx = pretend.stub(__call__=lambda a: context)

        monkeypatch.setattr(wheel_abi, "InGenericPkgCtx", InGenericPkgCtx)
        monkeypatch.setattr(
            wheel_abi, "elf_is_python_extension", lambda fn, elf: (fn, elf)
        )
        monkeypatch.setattr(
            wheel_abi,
            "elf_file_filter",
            lambda fns: [(fn, pretend.stub()) for fn in fns],
        )

        with pytest.raises(RuntimeError) as exec_info:
            wheel_abi.get_wheel_elfdata("/fakepath")

        assert exec_info.value.args == (message,)



# Sample numpy program that requires some BLAS and LAPACK routines to work
# properly
import numpy as np

rng = np.random.RandomState(0)
X = rng.randn(500, 200)
XTX = np.dot(X.T, X)
U, S, VT = np.linalg.svd(XTX)
if all(S > 0):
    print("ok")
else:
    print("[ERROR] invalid singular values:", S)



from jsonschema import validate

from auditwheel.policy import (
    POLICY_PRIORITY_HIGHEST,
    POLICY_PRIORITY_LOWEST,
    _load_policy_schema,
    load_policies,
    versioned_symbols_policy,
)


def test_policy():
    policy = load_policies()
    policy_schema = _load_policy_schema()
    validate(policy, policy_schema)


def test_policy_checks_glibc():
    policy = versioned_symbols_policy({"some_library.so": {"GLIBC_2.17"}})
    assert policy > POLICY_PRIORITY_LOWEST
    policy = versioned_symbols_policy({"some_library.so": {"GLIBC_999"}})
    assert policy == POLICY_PRIORITY_LOWEST
    policy = versioned_symbols_policy({"some_library.so": {"OPENSSL_1_1_0"}})
    assert policy == POLICY_PRIORITY_HIGHEST
    policy = versioned_symbols_policy({"some_library.so": {"IAMALIBRARY"}})
    assert policy == POLICY_PRIORITY_HIGHEST



import platform
import subprocess
import sys
import zipfile
from argparse import Namespace
from datetime import datetime, timezone
from pathlib import Path
from unittest import mock
from unittest.mock import Mock

import pytest

from auditwheel import main_repair
from auditwheel.wheel_abi import analyze_wheel_abi

HERE = Path(__file__).parent.resolve()


@pytest.mark.skipif(platform.machine() != "x86_64", reason="only supported on x86_64")
@pytest.mark.parametrize(
    "file, external_libs",
    [
        ("cffi-1.5.0-cp27-none-linux_x86_64.whl", {"libffi.so.5"}),
        ("python_snappy-0.5.2-pp260-pypy_41-linux_x86_64.whl", {"libsnappy.so.1"}),
    ],
)
def test_analyze_wheel_abi(file, external_libs):
    winfo = analyze_wheel_abi(str(HERE / file))
    assert set(winfo.external_refs["manylinux_2_5_x86_64"]["libs"]) == external_libs


@pytest.mark.skipif(platform.machine() != "x86_64", reason="only supported on x86_64")
def test_analyze_wheel_abi_pyfpe():
    winfo = analyze_wheel_abi(str(HERE / "fpewheel-0.0.0-cp35-cp35m-linux_x86_64.whl"))
    assert (
        winfo.sym_tag == "manylinux_2_5_x86_64"
    )  # for external symbols, it could get manylinux1
    assert (
        winfo.pyfpe_tag == "linux_x86_64"
    )  # but for having the pyfpe reference, it gets just linux


@pytest.mark.skipif(platform.machine() != "x86_64", reason="only checked on x86_64")
def test_wheel_source_date_epoch(tmp_path, monkeypatch):
    wheel_build_path = tmp_path / "wheel"
    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "wheel",
            "--no-deps",
            "-w",
            wheel_build_path,
            HERE / "sample_extension",
        ],
        check=True,
    )

    wheel_path, *_ = list(wheel_build_path.glob("*.whl"))
    wheel_output_path = tmp_path / "out"
    args = Namespace(
        LIB_SDIR=".libs",
        ONLY_PLAT=False,
        PLAT="manylinux_2_5_x86_64",
        STRIP=False,
        UPDATE_TAGS=True,
        WHEEL_DIR=str(wheel_output_path),
        WHEEL_FILE=[str(wheel_path)],
        EXCLUDE=[],
        cmd="repair",
        func=Mock(),
        prog="auditwheel",
        verbose=1,
    )
    monkeypatch.setenv("SOURCE_DATE_EPOCH", "650203200")
    # patchelf might not be available as we aren't running in a manylinux container
    # here. We don't need need it in this test, so just patch it.
    with mock.patch("auditwheel.patcher._verify_patchelf"):
        main_repair.execute(args, Mock())

    output_wheel, *_ = list(wheel_output_path.glob("*.whl"))
    with zipfile.ZipFile(output_wheel) as wheel_file:
        for file in wheel_file.infolist():
            assert (
                datetime(*file.date_time, tzinfo=timezone.utc).timestamp() == 650203200
            )



from unittest.mock import Mock, patch

import pytest
from elftools.common.exceptions import ELFError

from auditwheel.elfutils import (
    elf_file_filter,
    elf_find_ucs2_symbols,
    elf_find_versioned_symbols,
    elf_read_dt_needed,
    elf_references_PyFPE_jbuf,
)


class MockSymbol(dict):
    """Mock representing a Symbol in ELFTools."""

    def __init__(self, name, **kwargs):
        super().__init__(**kwargs)
        self._name = name

    @property
    def name(self):
        return self._name


@patch("auditwheel.elfutils.open")
@patch("auditwheel.elfutils.ELFFile")
class TestElfReadDt:
    def test_missing_section(self, elffile_mock, open_mock):
        # GIVEN
        open_mock.return_value.__enter__.return_value = Mock()
        elffile_mock.return_value.get_section_by_name.return_value = None

        # THEN
        with pytest.raises(ValueError):
            # WHEN
            elf_read_dt_needed("/fake.so")

    def test_needed_libs(self, elffile_mock, open_mock):
        # GIVEN
        open_mock.return_value.__enter__.return_value = Mock()
        section_mock = Mock()
        tag1 = Mock(needed="libz.so")
        tag1.entry.d_tag = "DT_NEEDED"
        tag2 = Mock(needed="libfoo.so")
        tag2.entry.d_tag = "DT_NEEDED"
        section_mock.iter_tags.return_value = [tag1, tag2]
        elffile_mock.return_value.get_section_by_name.return_value = section_mock

        # WHEN
        needed = elf_read_dt_needed("/fake.so")

        # THEN
        assert len(needed) == 2
        assert "libz.so" in needed
        assert "libfoo.so" in needed


@patch("auditwheel.elfutils.open")
@patch("auditwheel.elfutils.ELFFile")
class TestElfFileFilter:
    def test_filter(self, elffile_mock, open_mock):
        result = elf_file_filter(["file1.so", "file2.so"])
        assert len(list(result)) == 2

    def test_some_py_files(self, elffile_mock, open_mock):
        result = elf_file_filter(["file1.py", "file2.so", "file3.py"])
        assert len(list(result)) == 1

    def test_not_elf(self, elffile_mock, open_mock):
        # GIVEN
        elffile_mock.side_effect = ELFError

        # WHEN
        result = elf_file_filter(["file1.notelf", "file2.notelf"])

        # THEN
        assert len(list(result)) == 0


class TestElfFindVersionedSymbols:
    def test_find_symbols(self):
        # GIVEN
        elf = Mock()
        verneed = Mock()
        verneed.configure_mock(name="foo-lib")
        veraux = Mock()
        veraux.configure_mock(name="foo-lib")
        elf.get_section_by_name.return_value.iter_versions.return_value = (
            (verneed, [veraux]),
        )

        # WHEN
        symbols = list(elf_find_versioned_symbols(elf))

        # THEN
        assert symbols == [("foo-lib", "foo-lib")]

    @pytest.mark.parametrize("ld_name", ["ld-linux", "ld64.so.2", "ld64.so.1"])
    def test_only_ld_linux(self, ld_name):
        # GIVEN
        elf = Mock()
        verneed = Mock()
        verneed.configure_mock(name=ld_name)
        veraux = Mock()
        veraux.configure_mock(name="foo-lib")
        elf.get_section_by_name.return_value.iter_versions.return_value = (
            (verneed, [veraux]),
        )

        # WHEN
        symbols = list(elf_find_versioned_symbols(elf))

        # THEN
        assert len(symbols) == 0

    def test_empty_section(self):
        # GIVEN
        elf = Mock()
        elf.get_section_by_name.return_value = None

        # WHEN
        symbols = list(elf_find_versioned_symbols(elf))

        # THEN
        assert len(symbols) == 0


class TestFindUcs2Symbols:
    def test_elf_find_ucs2_symbols(self):
        # GIVEN
        elf = Mock()

        asunicode = MockSymbol(
            "PyUnicodeUCS2_AsUnicode",
            st_shndx="SHN_UNDEF",
            st_info=dict(type="STT_FUNC"),
        )
        symbols = (asunicode, Mock())
        symbols[1].name = "foobar"
        elf.get_section_by_name.return_value.iter_symbols.return_value = symbols

        # WHEN
        symbols = list(elf_find_ucs2_symbols(elf))

        # THEN
        assert len(symbols) == 1
        assert symbols[0] == "PyUnicodeUCS2_AsUnicode"

    def test_elf_find_ucs2_symbols_no_symbol(self):
        # GIVEN
        elf = Mock()

        symbols = (MockSymbol("FooSymbol"),)
        elf.get_section_by_name.return_value.iter_symbols.return_value = symbols

        # WHEN/THEN
        symbols = list(elf_find_ucs2_symbols(elf))
        assert len(symbols) == 0


class TestElfReferencesPyPFE:
    def test_elf_references_pyfpe_jbuf(self):
        # GIVEN
        elf = Mock()
        symbols = (
            MockSymbol(
                "PyFPE_jbuf", st_shndx="SHN_UNDEF", st_info=dict(type="STT_FUNC")
            ),
        )

        elf.get_section_by_name.return_value.iter_symbols.return_value = symbols

        # WHEN/THEN
        assert elf_references_PyFPE_jbuf(elf) is True

    def test_elf_references_pyfpe_jbuf_false(self):
        # GIVEN
        elf = Mock()
        symbols = (
            MockSymbol(
                "SomeSymbol", st_shndx="SHN_UNDEF", st_info=dict(type="STT_FUNC")
            ),
        )

        elf.get_section_by_name.return_value.iter_symbols.return_value = symbols

        # WHEN/THEN
        assert elf_references_PyFPE_jbuf(elf) is False

    def test_elf_references_pyfpe_jbuf_no_section(self):
        # GIVEN
        elf = Mock()

        # WHEN
        elf.get_section_by_name.return_value = None

        # WHEN/THEN
        assert elf_references_PyFPE_jbuf(elf) is False



import hashlib
from typing import BinaryIO


def hashfile(afile: BinaryIO, blocksize: int = 65536) -> str:
    """Hash the contents of an open file handle with SHA256"""
    hasher = hashlib.sha256()
    buf = afile.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = afile.read(blocksize)
    return hasher.hexdigest()



import os
from unittest.mock import call, patch

from auditwheel.patcher import Patchelf
from auditwheel.repair import append_rpath_within_wheel


@patch("auditwheel.patcher._verify_patchelf")
@patch("auditwheel.patcher.check_output")
@patch("auditwheel.patcher.check_call")
class TestRepair:
    def test_append_rpath(self, check_call, check_output, _):
        patcher = Patchelf()
        # When a library has an existing RPATH entry within wheel_dir
        existing_rpath = b"$ORIGIN/.existinglibdir"
        check_output.return_value = existing_rpath
        wheel_dir = "."
        lib_name = "test.so"
        full_lib_name = os.path.abspath(lib_name)
        append_rpath_within_wheel(lib_name, "$ORIGIN/.lib", wheel_dir, patcher)
        check_output_expected_args = [
            call(["patchelf", "--print-rpath", full_lib_name])
        ]
        # Then that entry is preserved when updating the RPATH
        check_call_expected_args = [
            call(["patchelf", "--remove-rpath", full_lib_name]),
            call(
                [
                    "patchelf",
                    "--force-rpath",
                    "--set-rpath",
                    f"{existing_rpath.decode()}:$ORIGIN/.lib",
                    full_lib_name,
                ]
            ),
        ]

        assert check_output.call_args_list == check_output_expected_args
        assert check_call.call_args_list == check_call_expected_args

    def test_append_rpath_reject_outside_wheel(self, check_call, check_output, _):
        patcher = Patchelf()
        # When a library has an existing RPATH entry outside wheel_dir
        existing_rpath = b"/outside/wheel/dir"
        check_output.return_value = existing_rpath
        wheel_dir = "/not/outside"
        lib_name = "test.so"
        full_lib_name = os.path.abspath(lib_name)
        append_rpath_within_wheel(lib_name, "$ORIGIN/.lib", wheel_dir, patcher)
        check_output_expected_args = [
            call(["patchelf", "--print-rpath", full_lib_name])
        ]
        # Then that entry is eliminated when updating the RPATH
        check_call_expected_args = [
            call(["patchelf", "--remove-rpath", full_lib_name]),
            call(
                [
                    "patchelf",
                    "--force-rpath",
                    "--set-rpath",
                    "$ORIGIN/.lib",
                    full_lib_name,
                ]
            ),
        ]

        assert check_output.call_args_list == check_output_expected_args
        assert check_call.call_args_list == check_call_expected_args

    def test_append_rpath_ignore_duplicates(self, check_call, check_output, _):
        patcher = Patchelf()
        # When a library has an existing RPATH entry and we try and append it again
        existing_rpath = b"$ORIGIN"
        check_output.return_value = existing_rpath
        wheel_dir = "."
        lib_name = "test.so"
        full_lib_name = os.path.abspath(lib_name)
        append_rpath_within_wheel(lib_name, "$ORIGIN", wheel_dir, patcher)
        check_output_expected_args = [
            call(["patchelf", "--print-rpath", full_lib_name])
        ]
        # Then that entry is ignored when updating the RPATH
        check_call_expected_args = [
            call(["patchelf", "--remove-rpath", full_lib_name]),
            call(
                ["patchelf", "--force-rpath", "--set-rpath", "$ORIGIN", full_lib_name]
            ),
        ]

        assert check_output.call_args_list == check_output_expected_args
        assert check_call.call_args_list == check_call_expected_args

    def test_append_rpath_ignore_relative(self, check_call, check_output, _):
        patcher = Patchelf()
        # When a library has an existing RPATH entry but it cannot be resolved
        # to an absolute path, it is eliminated
        existing_rpath = b"not/absolute"
        check_output.return_value = existing_rpath
        wheel_dir = "."
        lib_name = "test.so"
        full_lib_name = os.path.abspath(lib_name)
        append_rpath_within_wheel(lib_name, "$ORIGIN", wheel_dir, patcher)
        check_output_expected_args = [
            call(["patchelf", "--print-rpath", full_lib_name])
        ]
        # Then that entry is ignored when updating the RPATH
        check_call_expected_args = [
            call(["patchelf", "--remove-rpath", full_lib_name]),
            call(
                ["patchelf", "--force-rpath", "--set-rpath", "$ORIGIN", full_lib_name]
            ),
        ]

        assert check_output.call_args_list == check_output_expected_args
        assert check_call.call_args_list == check_call_expected_args



from unittest.mock import Mock, patch

from auditwheel.condatools import InCondaPkg, InCondaPkgCtx


@patch("auditwheel.condatools.tarbz2todir")
def test_in_condapkg(tarbz2todir_mock):
    with InCondaPkg("/fakepath"):
        assert True


@patch("auditwheel.condatools.tarbz2todir")
@patch("auditwheel.condatools.open")
def test_in_condapkg_context(open_mock, tarbz2todir_mock):
    with InCondaPkgCtx("/fakepath") as conda_pkg:
        file_mock = Mock()
        file_mock.readlines.return_value = ["file1\n", "file2\n", "\n"]
        open_mock.return_value.__enter__.return_value = file_mock
        # This returns empty lines so we have count with those as well. This
        # might be a subtle bug in the implementation.
        files = conda_pkg.iter_files()
        assert len(files) == 3
        assert "file1" and "file2" in files



import argparse
import os
import subprocess
import zipfile
from datetime import datetime, timezone
from typing import Any, Iterable, List, Optional


def unique_by_index(sequence: Iterable[Any]) -> List[Any]:
    """unique elements in `sequence` in the order in which they occur

    Parameters
    ----------
    sequence : iterable

    Returns
    -------
    uniques : list
        unique elements of sequence, ordered by the order in which the element
        occurs in `sequence`
    """
    uniques = []
    for element in sequence:
        if element not in uniques:
            uniques.append(element)
    return uniques


def zip2dir(zip_fname: str, out_dir: str) -> None:
    """Extract `zip_fname` into output directory `out_dir`

    Parameters
    ----------
    zip_fname : str
        Filename of zip archive to write
    out_dir : str
        Directory path containing files to go in the zip archive
    """
    with zipfile.ZipFile(zip_fname, "r") as z:
        for name in z.namelist():
            member = z.getinfo(name)
            extracted_path = z.extract(member, out_dir)
            attr = member.external_attr >> 16
            if member.is_dir():
                # this is always rebuilt as 755 by dir2zip
                os.chmod(extracted_path, 0o755)
            elif attr != 0:
                attr &= 511  # only keep permission bits
                attr |= 6 << 6  # at least read/write for current user
                os.chmod(extracted_path, attr)


def dir2zip(in_dir: str, zip_fname: str, date_time: Optional[datetime] = None) -> None:
    """Make a zip file `zip_fname` with contents of directory `in_dir`

    The recorded filenames are relative to `in_dir`, so doing a standard zip
    unpack of the resulting `zip_fname` in an empty directory will result in
    the original directory contents.

    Parameters
    ----------
    in_dir : str
        Directory path containing files to go in the zip archive
    zip_fname : str
        Filename of zip archive to write
    date_time : Optional[datetime]
        Time stamp to set on each file in the archive
    """
    if date_time is None:
        st = os.stat(in_dir)
        date_time = datetime.fromtimestamp(st.st_mtime, tz=timezone.utc)
    date_time_args = date_time.timetuple()[:6]
    compression = zipfile.ZIP_DEFLATED
    with zipfile.ZipFile(zip_fname, "w", compression=compression) as z:
        for root, dirs, files in os.walk(in_dir):
            for dir in dirs:
                dname = os.path.join(root, dir)
                out_dname = os.path.relpath(dname, in_dir) + "/"
                zinfo = zipfile.ZipInfo.from_file(dname, out_dname)
                zinfo.date_time = date_time_args
                z.writestr(zinfo, b"")
            for file in files:
                fname = os.path.join(root, file)
                out_fname = os.path.relpath(fname, in_dir)
                zinfo = zipfile.ZipInfo.from_file(fname, out_fname)
                zinfo.date_time = date_time_args
                zinfo.compress_type = compression
                with open(fname, "rb") as fp:
                    z.writestr(zinfo, fp.read())


def tarbz2todir(tarbz2_fname: str, out_dir: str) -> None:
    """Extract `tarbz2_fname` into output directory `out_dir`"""
    subprocess.check_output(["tar", "xjf", tarbz2_fname, "-C", out_dir])


class EnvironmentDefault(argparse.Action):
    """Get values from environment variable."""

    def __init__(self, env, required=True, default=None, **kwargs):
        self.env_default = os.environ.get(env)
        self.env = env
        if self.env_default:
            default = self.env_default
        if default:
            required = False
        if self.env_default and "choices" in kwargs:
            choices = kwargs["choices"]
            if self.env_default not in choices:
                self.option_strings = kwargs["option_strings"]
                args = {
                    "value": self.env_default,
                    "choices": ", ".join(map(repr, choices)),
                    "env": self.env,
                }
                msg = (
                    "invalid choice: %(value)r from environment variable "
                    "%(env)r (choose from %(choices)s)"
                )
                raise argparse.ArgumentError(self, msg % args)

        super().__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)



import argparse
import lzma
from pathlib import Path

import pytest

from auditwheel.tools import EnvironmentDefault, dir2zip, zip2dir


@pytest.mark.parametrize(
    ("environ", "passed", "expected"),
    [
        (None, None, "manylinux1"),
        (None, "manylinux2010", "manylinux2010"),
        ("manylinux2010", None, "manylinux2010"),
        ("manylinux2010", "linux", "linux"),
    ],
)
def test_environment_action(monkeypatch, environ, passed, expected):
    choices = ["linux", "manylinux1", "manylinux2010"]
    argv = []
    if passed:
        argv = ["--plat", passed]
    if environ:
        monkeypatch.setenv("AUDITWHEEL_PLAT", environ)
    p = argparse.ArgumentParser()
    p.add_argument(
        "--plat",
        action=EnvironmentDefault,
        env="AUDITWHEEL_PLAT",
        dest="PLAT",
        choices=choices,
        default="manylinux1",
    )
    args = p.parse_args(argv)
    assert args.PLAT == expected


def test_environment_action_invalid_env(monkeypatch):
    choices = ["linux", "manylinux1", "manylinux2010"]
    monkeypatch.setenv("AUDITWHEEL_PLAT", "foo")
    with pytest.raises(argparse.ArgumentError):
        p = argparse.ArgumentParser()
        p.add_argument(
            "--plat",
            action=EnvironmentDefault,
            env="AUDITWHEEL_PLAT",
            dest="PLAT",
            choices=choices,
            default="manylinux1",
        )


def _write_test_permissions_zip(path):
    source_zip_xz = Path(__file__).parent / "test-permissions.zip.xz"
    with lzma.open(source_zip_xz) as f:
        path.write_bytes(f.read())


def _check_permissions(path):
    for i in range(8):
        for j in range(8):
            for k in range(8):
                mode = (path / f"{i}{j}{k}.f").stat().st_mode
                assert ((mode >> 6) & 7) == (i | 6)  # always read/write
                assert ((mode >> 3) & 7) == j
                assert ((mode >> 0) & 7) == k
                mode = (path / f"{i}{j}{k}.d").stat().st_mode
                assert ((mode >> 6) & 7) == 7  # always read/write/execute
                assert ((mode >> 3) & 7) == 5  # always read/execute
                assert ((mode >> 0) & 7) == 5  # always read/execute


def test_zip2dir_permissions(tmp_path):
    source_zip = tmp_path / "test-permissions.zip"
    _write_test_permissions_zip(source_zip)
    extract_path = tmp_path / "unzip"
    zip2dir(str(source_zip), str(extract_path))
    _check_permissions(extract_path)


def test_zip2dir_round_trip_permissions(tmp_path):
    source_zip = tmp_path / "test-permissions.zip"
    _write_test_permissions_zip(source_zip)
    extract_path = tmp_path / "unzip2"
    zip2dir(str(source_zip), str(tmp_path / "unzip1"))
    dir2zip(str(tmp_path / "unzip1"), str(tmp_path / "tmp.zip"))
    zip2dir(str(tmp_path / "tmp.zip"), str(extract_path))
    _check_permissions(extract_path)


def test_dir2zip_deflate(tmp_path):
    buffer = b"\0" * 1024 * 1024
    input_dir = tmp_path / "input_dir"
    input_dir.mkdir()
    input_file = input_dir / "zeros.bin"
    input_file.write_bytes(buffer)
    output_file = tmp_path / "ouput.zip"
    dir2zip(str(input_dir), str(output_file))
    assert output_file.stat().st_size < len(buffer) / 4



import re
from distutils.spawn import find_executable
from itertools import chain
from subprocess import CalledProcessError, check_call, check_output
from typing import Tuple


class ElfPatcher:
    def replace_needed(self, file_name: str, *old_new_pairs: Tuple[str, str]) -> None:
        raise NotImplementedError

    def set_soname(self, file_name: str, new_so_name: str) -> None:
        raise NotImplementedError

    def set_rpath(self, file_name: str, rpath: str) -> None:
        raise NotImplementedError

    def get_rpath(self, file_name: str) -> str:
        raise NotImplementedError


def _verify_patchelf() -> None:
    """This function looks for the ``patchelf`` external binary in the PATH,
    checks for the required version, and throws an exception if a proper
    version can't be found. Otherwise, silcence is golden
    """
    if not find_executable("patchelf"):
        raise ValueError("Cannot find required utility `patchelf` in PATH")
    try:
        version = check_output(["patchelf", "--version"]).decode("utf-8")
    except CalledProcessError:
        raise ValueError("Could not call `patchelf` binary")

    m = re.match(r"patchelf\s+(\d+(.\d+)?)", version)
    if m and tuple(int(x) for x in m.group(1).split(".")) >= (0, 14):
        return
    raise ValueError(
        f"patchelf {version} found. auditwheel repair requires " "patchelf >= 0.14."
    )


class Patchelf(ElfPatcher):
    def __init__(self) -> None:
        _verify_patchelf()

    def replace_needed(self, file_name: str, *old_new_pairs: Tuple[str, str]) -> None:
        check_call(
            [
                "patchelf",
                *chain.from_iterable(
                    ("--replace-needed", *pair) for pair in old_new_pairs
                ),
                file_name,
            ]
        )

    def set_soname(self, file_name: str, new_so_name: str) -> None:
        check_call(["patchelf", "--set-soname", new_so_name, file_name])

    def set_rpath(self, file_name: str, rpath: str) -> None:

        check_call(["patchelf", "--remove-rpath", file_name])
        check_call(["patchelf", "--force-rpath", "--set-rpath", rpath, file_name])

    def get_rpath(self, file_name: str) -> str:

        return (
            check_output(["patchelf", "--print-rpath", file_name])
            .decode("utf-8")
            .strip()
        )



import os
import subprocess

from setuptools import Extension, setup

cmd = "gcc -fPIC -shared -o b/libb.so b/b.c"
subprocess.check_call(cmd.split())
cmd = (
    "gcc -fPIC -shared -o a/liba.so "
    "-Wl,{dtags_flag} -Wl,-rpath=$ORIGIN/../b "
    "-Ib a/a.c -Lb -lb"
).format(
    dtags_flag=(
        "--enable-new-dtags"
        if os.getenv("DTAG") == "runpath"
        else "--disable-new-dtags"
    )
)
subprocess.check_call(cmd.split())

setup(
    name="testrpath",
    version="0.0.1",
    packages=["testrpath"],
    package_dir={"": "src"},
    ext_modules=[
        Extension(
            "testrpath/testrpath",
            sources=["src/testrpath/testrpath.c"],
            include_dirs=["a"],
            libraries=["a"],
            library_dirs=["a"],
        )
    ],
)



import sys

import pytest

from auditwheel.main import main

on_supported_platform = pytest.mark.skipif(
    sys.platform != "linux", reason="requires Linux system"
)


def test_unsupported_platform(monkeypatch):
    # GIVEN
    monkeypatch.setattr(sys, "platform", "unsupported_platform")

    # WHEN
    retval = main()

    # THEN
    assert retval == 1


@on_supported_platform
def test_help(monkeypatch, capsys):
    # GIVEN
    monkeypatch.setattr(sys, "argv", ["auditwheel"])

    # WHEN
    retval = main()

    # THEN
    assert retval is None
    captured = capsys.readouterr()
    assert "usage: auditwheel [-h] [-V] [-v] command ..." in captured.out



from io import BytesIO

from auditwheel.hashfile import hashfile


def test_hash():
    # GIVEN
    mock_file = BytesIO(b"this is a test file")

    # WHEN
    result = hashfile(mock_file)

    # THEN
    assert result == "5881707e54b0112f901bc83a1ffbacac8fab74ea46a6f706a3efc5f7d4c1c625"



import logging
from enum import IntEnum

from .error import InvalidLibc
from .musllinux import find_musl_libc

logger = logging.getLogger(__name__)


class Libc(IntEnum):
    GLIBC = (1,)
    MUSL = (2,)


def get_libc() -> Libc:
    try:
        find_musl_libc()
        logger.debug("Detected musl libc")
        return Libc.MUSL
    except InvalidLibc:
        logger.debug("Falling back to GNU libc")
        return Libc.GLIBC



import logging
from collections import OrderedDict

logger = logging.getLogger(__name__)


def configure_parser(sub_parsers):
    help = "Audit a wheel for external shared library dependencies."
    p = sub_parsers.add_parser("show", help=help, description=help)
    p.add_argument("WHEEL_FILE", help="Path to wheel file.")
    p.set_defaults(func=execute)


def printp(text: str) -> None:
    from textwrap import wrap

    print()
    print("\n".join(wrap(text, break_long_words=False, break_on_hyphens=False)))


def execute(args, p):
    import json
    from os.path import basename, isfile

    from .policy import (
        POLICY_PRIORITY_HIGHEST,
        POLICY_PRIORITY_LOWEST,
        get_policy_name,
        get_priority_by_name,
        load_policies,
    )
    from .wheel_abi import NonPlatformWheel, analyze_wheel_abi

    fn = basename(args.WHEEL_FILE)

    if not isfile(args.WHEEL_FILE):
        p.error("cannot access %s. No such file" % args.WHEEL_FILE)

    try:
        winfo = analyze_wheel_abi(args.WHEEL_FILE)
    except NonPlatformWheel:
        logger.info(NonPlatformWheel.LOG_MESSAGE)
        return 1

    libs_with_versions = [
        f"{k} with versions {v}" for k, v in winfo.versioned_symbols.items()
    ]

    printp(
        '%s is consistent with the following platform tag: "%s".'
        % (fn, winfo.overall_tag)
    )

    if get_priority_by_name(winfo.pyfpe_tag) < POLICY_PRIORITY_HIGHEST:
        printp(
            "This wheel uses the PyFPE_jbuf function, which is not compatible with the"
            " manylinux1 tag. (see https://www.python.org/dev/peps/pep-0513/"
            "#fpectl-builds-vs-no-fpectl-builds)"
        )
        if args.verbose < 1:
            return

    if get_priority_by_name(winfo.ucs_tag) < POLICY_PRIORITY_HIGHEST:
        printp(
            "This wheel is compiled against a narrow unicode (UCS2) "
            "version of Python, which is not compatible with the "
            "manylinux1 tag."
        )
        if args.verbose < 1:
            return

    if len(libs_with_versions) == 0:
        printp(
            "The wheel references no external versioned symbols from "
            "system-provided shared libraries."
        )
    else:
        printp(
            "The wheel references external versioned symbols in these "
            "system-provided shared libraries: %s" % ", ".join(libs_with_versions)
        )

    if get_priority_by_name(winfo.sym_tag) < POLICY_PRIORITY_HIGHEST:
        printp(
            (
                'This constrains the platform tag to "%s". '
                "In order to achieve a more compatible tag, you would "
                "need to recompile a new wheel from source on a system "
                "with earlier versions of these libraries, such as "
                "a recent manylinux image."
            )
            % winfo.sym_tag
        )
        if args.verbose < 1:
            return

    libs = winfo.external_refs[get_policy_name(POLICY_PRIORITY_LOWEST)]["libs"]
    if len(libs) == 0:
        printp("The wheel requires no external shared libraries! :)")
    else:
        printp("The following external shared libraries are required " "by the wheel:")
        print(json.dumps(OrderedDict(sorted(libs.items())), indent=4))

    for p in sorted(load_policies(), key=lambda p: p["priority"]):
        if p["priority"] > get_priority_by_name(winfo.overall_tag):
            libs = winfo.external_refs[p["name"]]["libs"]
            if len(libs):
                printp(
                    (
                        'In order to achieve the tag platform tag "%s" '
                        "the following shared library dependencies "
                        "will need to be eliminated:"
                    )
                    % p["name"]
                )
                printp(", ".join(sorted(libs.keys())))
            blacklist = winfo.external_refs[p["name"]]["blacklist"]
            if len(blacklist):
                printp(
                    (
                        'In order to achieve the tag platform tag "%s" '
                        "the following black-listed symbol dependencies "
                        "will need to be eliminated:"
                    )
                    % p["name"]
                )
                for key in sorted(blacklist.keys()):
                    printp(f"From {key}: " + ", ".join(sorted(blacklist[key])))
